# Copyright (c) OpenMMLab. All rights reserved.
from typing import Sequence

from torch.utils.data import BatchSampler, Sampler

from mmtrack.registry import DATA_SAMPLERS


@DATA_SAMPLERS.register_module()
class EntireVideoBatchSampler(BatchSampler):
    """A sampler wrapper for grouping images from one video into a same batch.

    Args:
        sampler (Sampler): Base sampler.
        batch_size (int): Size of mini-batch. Here, we take a video as a batch.
            Defaults to 1.
        drop_last (bool): If ``True``, the sampler will drop the last batch if
            its size would be less than ``batch_size``. Defaults to False.
    """

    def __init__(self,
                 sampler: Sampler,
                 batch_size: int = 1,
                 drop_last: bool = False) -> None:
        assert sampler.dataset.test_mode
        if not isinstance(sampler, Sampler):
            raise TypeError('sampler should be an instance of ``Sampler``, '
                            f'but got {sampler}')
        if not isinstance(batch_size, int) or batch_size != 1:
            raise ValueError('batch_size should be 1, '
                             f'but got batch_size={batch_size}')
        self.sampler = sampler
        self.batch_size = batch_size
        self.drop_last = drop_last

    def __iter__(self) -> Sequence[int]:
        batch = []
        for idx in self.sampler:
            data_info = self.sampler.dataset.get_data_info(idx)
            video_length = data_info['video_length']
            batch.append(idx)
            if len(batch) == video_length:
                yield batch
                batch = []

    def __len__(self) -> int:
        return self.sampler.num_videos



_base_ = [
    './tracktor_faster-rcnn_r50-fpn_8xb2-4e_mot17halftrain'
    '_test-mot17halfval.py'
]

model = dict(
    tracker=dict(
        type='TracktorTracker',
        obj_score_thr=[0.4, 0.5, 0.6],
        regression=dict(
            obj_score_thr=[0.4, 0.5, 0.6],
            nms=dict(type='nms', iou_threshold=0.6),
            match_iou_thr=[0.3, 0.5]),
        reid=dict(
            num_samples=10,
            img_scale=(256, 128),
            img_norm_cfg=None,
            match_score_thr=2.0,
            match_iou_thr=0.2),
        momentums=None,
        num_frames_retain=10))



# Copyright (c) OpenMMLab. All rights reserved.
from .formatting import CheckPadMaskValidity, PackReIDInputs, PackTrackInputs
from .loading import LoadTrackAnnotations
from .processing import PairSampling, TridentSampling
from .transforms import (BrightnessAug, CropLikeDiMP, CropLikeSiamFC, GrayAug,
                         SeqBboxJitter, SeqBlurAug, SeqColorAug,
                         SeqCropLikeStark, SeqShiftScaleAug)

__all__ = [
    'LoadTrackAnnotations', 'PackTrackInputs', 'PackReIDInputs',
    'PairSampling', 'CropLikeSiamFC', 'SeqShiftScaleAug', 'SeqColorAug',
    'SeqBlurAug', 'TridentSampling', 'GrayAug', 'BrightnessAug',
    'SeqBboxJitter', 'SeqCropLikeStark', 'CheckPadMaskValidity', 'CropLikeDiMP'
]



# Copyright (c) OpenMMLab. All rights reserved.
import math
from typing import List, Optional, Tuple

import torch
import torch.nn.functional as F
from mmcv.image import imrotate
from mmdet.structures.bbox import bbox_cxcywh_to_xyxy, bbox_xyxy_to_cxcywh
from mmengine.structures import InstanceData
from torch import Tensor

from mmtrack.registry import MODELS
from mmtrack.utils import (InstanceList, OptConfigType, OptMultiConfig,
                           SampleList, gauss_blur)
from .base import BaseSingleObjectTracker


@MODELS.register_module()
class PrDiMP(BaseSingleObjectTracker):
    """PrDiMP: Probabilistic Regression for Visual Tracking.

    This single object tracker is the implementation of `PrDiMP
    <https://arxiv.org/abs/2003.12565>`_.

    args:
        backbone (dict, optional): the configuration of backbone network.
            Defaults to None.
        cls_head (dict, optional):  target classification module.
            Defaults to None.
        bbox_head (dict, optional):  bounding box regression module.
            Defaults to None.
        init_cfg (dict, optional): the configuration of initialization.
            Defaults to None.
        test_cfg (dict, optional): the configuration of test.
            Defaults to None.

    """

    def __init__(self,
                 backbone: dict,
                 neck: Optional[dict] = None,
                 cls_head: Optional[dict] = None,
                 bbox_head: Optional[dict] = None,
                 train_cfg: Optional[dict] = None,
                 test_cfg: Optional[dict] = None,
                 data_preprocessor: OptConfigType = None,
                 init_cfg: OptMultiConfig = None):
        super(PrDiMP, self).__init__(data_preprocessor, init_cfg)
        self.backbone = MODELS.build(backbone)
        cls_head.update(test_cfg=test_cfg)
        bbox_head.update(test_cfg=test_cfg)
        self.classifier = MODELS.build(cls_head)
        self.bbox_regressor = MODELS.build(bbox_head)
        self.test_cfg = test_cfg

    def init(self, img: Tensor) -> Tuple[Tuple[Tensor, ...], Tensor]:
        """Initialize tracker.

        Args:
            img (Tensor): Input image of shape (1, C, H, W).
            init_bbox (Tensor): of (4, ) shape in [cx, cy, w, h] format.
        """
        self.frame_num = 1
        init_bbox = self.memo.bbox

        # Set size for image and cropped sample. img_size is in [w, h] format.
        self.img_size = torch.Tensor([img.shape[-1],
                                      img.shape[-2]]).to(init_bbox.device)
        sample_size = self.test_cfg['img_sample_size']
        sample_size = torch.Tensor([sample_size, sample_size] if isinstance(
            sample_size, int) else sample_size)
        self.sample_size = sample_size.to(init_bbox.device)

        # Compute expanded size and output size about augmentation
        aug_expansion_factor = self.test_cfg['init_aug_cfg'][
            'aug_expansion_factor']
        aug_expansion_size = (self.sample_size * aug_expansion_factor).float()

        # Crop image patches and bboxes
        img_patch, patch_coord = self.get_cropped_img(
            img, init_bbox.round(),
            self.test_cfg['search_scale_factor'] * aug_expansion_factor,
            aug_expansion_size)
        resized_factor = (patch_coord[:, 2:4] /
                          aug_expansion_size).prod().sqrt()
        init_bbox = self.generate_bbox(init_bbox, init_bbox[:2].round(),
                                       resized_factor)

        # Crop patches from the image and perform augmentation on the image
        # patches
        aug_img_patches, aug_cls_bboxes = self.gen_aug_imgs_bboxes(
            img_patch, init_bbox, self.sample_size)

        # `init_backbone_feats` is a tuple containing the features of `layer2`
        # and `layer3`
        init_backbone_feats = self.backbone(aug_img_patches)

        # Initialize the classifier with bboxes and features of `layer3`
        # get the augmented bboxes on the augmented image patches
        self.classifier.init_classifier(
            init_backbone_feats[-1], aug_cls_bboxes,
            self.test_cfg['init_aug_cfg']['augmentation']['dropout'])

        # Initialize IoUNet
        # only use the features of the non-augmented image
        init_iou_features = [x[:1] for x in init_backbone_feats]
        self.bbox_regressor.init_iou_net(init_iou_features, init_bbox)

    def img_shift_crop(self,
                       img: Tensor,
                       output_size: Optional[List] = None,
                       shift: Optional[List] = None) -> Tensor:
        """Shift and crop the image.

        Args:
            img (Tensor): The image of shape (C, H, W).
            output_size (list): in [w, h] format.
            shift (list): in [x, y] fotmat.

        Returns:
            Tensor: Augmented image.
        """
        img_size = [img.shape[-1], img.shape[-2]]
        # img_size = img.shape[-2:]
        if output_size is None:
            pad_h = 0
            pad_w = 0
        else:
            pad_w = (output_size[0] - img_size[0]) / 2
            pad_h = (output_size[1] - img_size[1]) / 2

        if shift is None:
            shift = [0, 0]

        pad_left = math.floor(pad_w) + shift[0]
        pad_right = math.ceil(pad_w) - shift[0]
        pad_top = math.floor(pad_h) + shift[1]
        pad_bottom = math.ceil(pad_h) - shift[1]

        return F.pad(img, (pad_left, pad_right, pad_top, pad_bottom),
                     'replicate')

    def gen_aug_imgs_bboxes(self, img: Tensor, init_bbox: Tensor,
                            output_size: Tensor) -> Tensor:
        """Perform data augmentation.

        Args:
            img (Tensor): Cropped image of shape (1, C, H, W).
            init_bbox (Tensor): of (4, ) shape in [cx, cy, w, h] format.
            output_size (Tensor): of (2, ) shape in [w, h] format.

        Returns:
            Tensor: The cropped augmented image patches.
        """
        output_size = output_size.long().cpu().tolist()
        random_shift_factor = self.test_cfg['init_aug_cfg'][
            'random_shift_factor']

        def get_rand_shift():
            return ((torch.rand(2) - 0.5) * self.sample_size.cpu() *
                    random_shift_factor).long().tolist()

        augs = self.test_cfg['init_aug_cfg']['augmentation']
        aug_imgs = [self.img_shift_crop(img, output_size)]
        aug_bboxes = [init_bbox]

        # All augmentations
        if 'relativeshift' in augs:
            for shift in augs['relativeshift']:
                absulute_shift = (torch.Tensor(shift) *
                                  self.sample_size.cpu() / 2).long().tolist()
                aug_imgs.append(
                    self.img_shift_crop(img, output_size, absulute_shift))
                bbox_shift = torch.tensor(
                    absulute_shift + [0, 0], device=init_bbox.device)
                aug_bboxes.append(init_bbox + bbox_shift)

        if 'fliplr' in augs and augs['fliplr']:
            shift = get_rand_shift()
            aug_imgs.append(
                self.img_shift_crop(img.flip(3), output_size, shift))
            bbox_shift = torch.tensor(shift + [0, 0], device=init_bbox.device)
            aug_bboxes.append(init_bbox + bbox_shift)

        if 'blur' in augs:
            for sigma in augs['blur']:
                kernel_size = [math.ceil(2 * s) for s in sigma]
                img_blur = gauss_blur(
                    img, kernel_size=kernel_size, sigma=sigma)
                shift = get_rand_shift()
                aug_imgs.append(
                    self.img_shift_crop(img_blur, output_size, shift))
                bbox_shift = torch.tensor(
                    shift + [0, 0], device=init_bbox.device)
                aug_bboxes.append(init_bbox + bbox_shift)

        if 'rotate' in augs:
            for angle in augs['rotate']:
                img_numpy = img.squeeze(0).permute(1, 2, 0).cpu().numpy()
                assert img_numpy.ndim == 3
                rotated_img = imrotate(
                    img_numpy, angle, border_mode='replicate')
                img_tensor = torch.from_numpy(rotated_img.transpose(
                    2, 0, 1)).float().unsqueeze(0).to(img.device)
                shift = get_rand_shift()
                aug_imgs.append(
                    self.img_shift_crop(img_tensor, output_size, shift))
                bbox_shift = torch.tensor(
                    shift + [0, 0], device=init_bbox.device)
                aug_bboxes.append(init_bbox + bbox_shift)

        if 'dropout' in augs:
            for _ in range(len(augs['dropout'])):
                aug_bboxes.append(init_bbox)

        aug_imgs = torch.cat(aug_imgs, dim=0)
        aug_bboxes = torch.stack(aug_bboxes)
        return aug_imgs, aug_bboxes

    def generate_bbox(self, bbox: Tensor, sample_center: Tensor,
                      resize_factor: float) -> Tensor:
        """All inputs are based in original image coordinates and the outputs
        are based on the resized cropped image sample.

        Args:
            bbox (Tensor): of shape (4,) in [cx, cy, w, h] format
            sample_center (Tensor): of shape (2,)
            resize_factor (float):

        Return:
            Tensor: in [cx, cy, w, h] format
        """
        bbox_center = (bbox[:2] - sample_center) / resize_factor + (
            self.sample_size / 2)
        bbox_size = bbox[2:4] / resize_factor
        return torch.cat([bbox_center, bbox_size])

    def track(self, img: Tensor, data_samples: SampleList) -> InstanceList:
        """Track the box `bbox` of previous frame to current frame `img`.

        Args:
            img (Tensor): of shape (1, C, H, W).
            bbox (list | Tensor): The bbox in previous frame. The shape of the
                bbox is (4, ) in [cx, cy, w, h] format.

        Returns:
            conf_score (int): The confidence of predicted bbox.
            bbox (Tensor): The predicted bbox in [cx, cy, w, h] format
        """
        self.frame_num += 1
        bbox = self.memo.bbox.clone()

        # 1. Extract backbone features
        img_patch, patch_coord = self.get_cropped_img(
            img,
            bbox.round(),
            self.test_cfg['search_scale_factor'],
            self.sample_size,
            border_mode=self.test_cfg['border_mode'],
            max_scale_change=self.test_cfg['patch_max_scale_change'])

        backbone_feats = self.backbone(img_patch)

        # location of sample
        sample_center = patch_coord[:, :2].squeeze()
        sample_scale_factor = (patch_coord[:, 2:] /
                               self.sample_size).prod(dim=1).sqrt()

        # 2. Locate the target roughly using score map.
        new_bbox_center, score_map, state = self.classifier.predict(
            backbone_feats, data_samples, bbox, sample_center,
            sample_scale_factor)

        # 3. Refine position and scale of the target.
        if state != 'not_found':
            inside_offset = (self.test_cfg['bbox_inside_ratio'] -
                             0.5) * bbox[2:4]
            # clip the coordinates of the center of the target on the original
            # image
            bbox[:2] = torch.max(
                torch.min(new_bbox_center, self.img_size - inside_offset),
                inside_offset)

            cls_bboxes = self.generate_bbox(bbox, sample_center,
                                            sample_scale_factor)
            new_bbox = self.bbox_regressor.predict(backbone_feats,
                                                   data_samples, cls_bboxes,
                                                   sample_center,
                                                   sample_scale_factor)
            if new_bbox is not None:
                bbox = new_bbox

        # 4. Update the classifier
        update_flag = state not in ['not_found', 'uncertain']
        # Update the classifier filter using the latest position and size of
        # target
        if update_flag:
            # Create the target_bbox using the refined predicted boxes
            target_bbox = self.generate_bbox(bbox, sample_center,
                                             sample_scale_factor)
            hard_neg_flag = (state == 'hard_negative')
            # Update the filter of classifier using it's optimizer module
            self.classifier.update_classifier(target_bbox, self.frame_num,
                                              hard_neg_flag)

        result = [InstanceData()]
        result[0].scores = torch.max(score_map[0]).unsqueeze(0)
        result[0].bboxes = bbox_cxcywh_to_xyxy(bbox.unsqueeze(0))

        return result

    def get_cropped_img(
            self,
            img: Tensor,
            target_bbox: Tensor,
            search_scale_factor: float,
            output_size: Optional[Tensor] = None,
            border_mode: str = 'replicate',
            max_scale_change: Optional[float] = None) -> Tuple[Tensor, Tensor]:
        """Get the cropped patch based on the original image.

        Args:
            img (Tensor): The original image.
            target_bbox (Tensor): The bbox of target in [cx, cy, w,h] format.
            search_scale_factor (float): The ratio of cropped size to the size
                of target bbox.
            output_size (Optional[Tensor], optional): The output size.
                Defaults to None.
            border_mode (str, optional): The border mode. Defaults to
                'replicate'.
            max_scale_change (Optional[float], optional): The max scale change.
                Defaults to None.
            is_mask (Optional[bool], optional): Whether is mask.
                Defaults to False.

        Returns:
            img_patch (Tensor): of (1, c, h, w) shape.
            patch_coord (Tensor): of (1, 4) shape in [cx, cy, w, h] format.
        """
        crop_size = target_bbox[2:4].prod().sqrt() * search_scale_factor

        # Get new sample size if forced inside the image
        if border_mode == 'inside' or border_mode == 'inside_major':
            img_sz = torch.Tensor([img.shape[3],
                                   img.shape[2]]).to(target_bbox.device)
            shrink_factor = (crop_size.float() / img_sz)
            if border_mode == 'inside':
                shrink_factor = shrink_factor.max()
            elif border_mode == 'inside_major':
                shrink_factor = shrink_factor.min()
            shrink_factor.clamp_(min=1, max=max_scale_change)
            crop_size = (crop_size.float() / shrink_factor).long()

        tl = (target_bbox[:2] - crop_size // 2).long()
        br = (target_bbox[:2] + crop_size // 2).long()

        # Shift the crop to inside
        if border_mode == 'inside' or border_mode == 'inside_major':
            img2_sz = torch.LongTensor([img.shape[3],
                                        img.shape[2]]).to(target_bbox.device)
            shift = (-tl).clamp(0) - (br - img2_sz).clamp(0)
            tl += shift
            br += shift

            outside = torch.floor_divide(
                ((-tl).clamp(0) + (br - img2_sz).clamp(0)), 2)
            shift = (-tl - outside) * (outside > 0).long()
            tl += shift
            br += shift

        patch_coord = torch.cat((tl, br)).view(1, 4)
        patch_coord = bbox_xyxy_to_cxcywh(patch_coord)

        # Crop image patch
        img_patch = F.pad(
            img, (-tl[0].item(), br[0].item() - img.shape[3], -tl[1].item(),
                  br[1].item() - img.shape[2]),
            mode='replicate')

        if output_size is None:
            return img_patch.clone(), patch_coord

        # Resize
        img_patch = F.interpolate(
            img_patch,
            output_size.long().flip(0).tolist(),
            mode='bilinear',
            align_corners=True)

        return img_patch, patch_coord

    def loss(self, batch_inputs: dict, batch_data_samples: SampleList,
             **kwargs) -> dict:
        """
        Args:
            batch_inputs (Dict[str, Tensor]): of shape (N, T, C, H, W) encoding
                input images. Typically these should be mean centered and std
                scaled. The N denotes batch size. The T denotes the number of
                key/reference frames.
                - img (Tensor) : The key images.
                - ref_img (Tensor): The reference images.

            batch_data_samples (list[:obj:`TrackDataSample`]): The batch
                data samples. It usually includes information such
                as ``gt_instance``.

        Return:
            dict: A dictionary of loss components.
        """
        search_img = batch_inputs['search_img']
        assert search_img.dim(
        ) == 5, 'The img must be 5D Tensor (N, T, C, H, W).'
        # has [T * N, C, H, W] shape and the first N images cover the entire
        # mini-batch.
        search_img = search_img.transpose(1, 0).contiguous().view(
            -1, *search_img.shape[2:])

        template_img = batch_inputs['img']
        assert template_img.dim(
        ) == 5, 'The img must be 5D Tensor (N, T, C, H, W).'
        # has [T * N, C, H, W] shape and the first N images cover the entire
        # mini-batch.
        template_img = template_img.transpose(1, 0).contiguous().view(
            -1, *template_img.shape[2:])

        z_feat = self.backbone(template_img)
        x_feat = self.backbone(search_img)

        losses = dict()
        loss_cls = self.classifier.loss(z_feat, x_feat, batch_data_samples,
                                        **kwargs)

        loss_bbox = self.bbox_regressor.loss(z_feat, x_feat,
                                             batch_data_samples, **kwargs)
        losses.update(loss_cls)
        losses.update(loss_bbox)
        return losses



_base_ = [
    '../../_base_/datasets/trackingnet.py',
    './stark-st2_r50_8xb16-50e_got10k-lasot-trackingnet-coco_base.py'
]

# model setting
model = dict(test_cfg=dict(update_intervals=[25]))

# evaluator
val_evaluator = dict(outfile_prefix='results/stark_st2_trackingnet')
test_evaluator = val_evaluator



# Copyright (c) OpenMMLab. All rights reserved.
from .aggregators import *  # noqa: F401,F403
from .backbones import *  # noqa: F401,F403
from .data_preprocessors import *  # noqa: F401,F403
from .filter import *  # noqa: F401,F403
from .layers import *  # noqa: F401,F403
from .losses import *  # noqa: F401,F403
from .mot import *  # noqa: F401,F403
from .motion import *  # noqa: F401,F403
from .reid import *  # noqa: F401,F403
from .roi_heads import *  # noqa: F401,F403
from .sot import *  # noqa: F401,F403
from .task_modules import *  # noqa: F401,F403
from .track_heads import *  # noqa: F401,F403
from .trackers import *  # noqa: F401,F403
from .vid import *  # noqa: F401,F403
from .vis import *  # noqa: F401,F403



# Copyright (c) OpenMMLab. All rights reserved.
import os
import time
from typing import List, Union

import numpy as np

from mmtrack.registry import DATASETS
from .base_sot_dataset import BaseSOTDataset


@DATASETS.register_module()
class TrackingNetDataset(BaseSOTDataset):
    """TrackingNet dataset of single object tracking.

    The dataset can both support training and testing mode.
    """

    def __init__(self,
                 chunks_list: List[Union[int, str]] = ['all'],
                 *args,
                 **kwargs):
        """Initialization of SOT dataset class.

        Args:
            chunks_list (list[Union[int, str]], optional): The training chunks.
                The optional values in this list are: 0, 1, 2, ..., 10, 11 and
                'all'. Some methods may only use part of the dataset. Default
                to all chunks, namely ['all'].
        """
        if isinstance(chunks_list, (str, int)):
            chunks_list = [chunks_list]
        assert set(chunks_list).issubset(set(range(12)) | {'all'})
        if 'all' in chunks_list:
            self.chunks_list = [f'TRAIN_{i}' for i in range(12)]
        else:
            self.chunks_list = [f'TRAIN_{chunk}' for chunk in chunks_list]
        super(TrackingNetDataset, self).__init__(*args, **kwargs)

    def load_data_list(self) -> List[dict]:
        """Load annotations from an annotation file named as ``self.ann_file``.

        Args:
            split (str, optional): the split of dataset. Defaults to 'train'.

        Returns:
            list[dict]: the length of the list is the number of videos. The
                inner dict is in the following format:
                    {
                        'video_path': the video path
                        'ann_path': the annotation path
                        'start_frame_id': the starting frame ID number
                            contained in the image name
                        'end_frame_id': the ending frame ID number contained in
                            the image name
                        'framename_template': the template of image name
                    }
        """
        print('Loading TrackingNet dataset...')
        start_time = time.time()
        chunks = set(self.chunks_list)
        data_infos = []
        data_infos_str = self._loadtxt(
            self.ann_file, return_ndarray=False).split('\n')
        # the first line of annotation file is a dataset comment.
        for line in data_infos_str[1:]:
            # compatible with different OS.
            line = line.strip().replace('/', os.sep).split(',')
            chunk = line[0].split(os.sep)[0]
            if chunk == 'TEST' or chunk in chunks:
                data_info = dict(
                    video_path=line[0],
                    ann_path=line[1],
                    start_frame_id=int(line[2]),
                    end_frame_id=int(line[3]),
                    framename_template='%d.jpg')
                data_infos.append(data_info)
        print(f'TrackingNet dataset loaded! ({time.time()-start_time:.2f} s)')
        return data_infos

    def prepare_test_data(self, video_idx: int, frame_idx: int) -> dict:
        """Get testing data of one frame. We parse one video, get one frame
        from it and pass the frame information to the pipeline.

        Args:
            video_idx (int): The index of video.
            frame_idx (int): The index of frame.

        Returns:
            dict: Testing data of one frame.
        """
        if self.test_memo.get('video_idx', None) != video_idx:
            self.test_memo.video_idx = video_idx
            ann_infos = self.get_ann_infos_from_video(video_idx)
            img_infos = self.get_img_infos_from_video(video_idx)
            self.test_memo.video_infos = dict(**img_infos, **ann_infos)
        assert 'video_idx' in self.test_memo and 'video_infos'\
            in self.test_memo

        results = {}
        results['img_path'] = self.test_memo.video_infos['img_paths'][
            frame_idx]
        results['frame_id'] = frame_idx
        results['video_id'] = video_idx
        results['video_length'] = self.test_memo.video_infos['video_length']

        instance = {}
        if frame_idx == 0:
            ann_infos = self.get_ann_infos_from_video(video_idx)
            instance['bbox'] = ann_infos['bboxes'][frame_idx]

        results['instances'] = []
        instance['visible'] = True
        instance['bbox_label'] = np.array([0], dtype=np.int32)
        results['instances'].append(instance)
        results = self.pipeline(results)
        return results



# Copyright (c) OpenMMLab. All rights reserved.
import argparse
import os.path as osp

import mmengine
import numpy as np
from mmdet.models.utils import mask2ndarray
from mmengine import Config, DictAction
from mmengine.structures import InstanceData

from mmtrack.registry import DATASETS, VISUALIZERS
from mmtrack.structures import TrackDataSample
from mmtrack.utils import register_all_modules


def parse_args():
    parser = argparse.ArgumentParser(description='Browse a dataset')
    parser.add_argument('config', help='train config file path')
    parser.add_argument(
        '--output-dir',
        default=None,
        type=str,
        help='If there is no display interface, you can save it')
    parser.add_argument('--not-show', default=False, action='store_true')
    parser.add_argument(
        '--show-interval',
        type=float,
        default=2,
        help='the interval of show (s)')
    parser.add_argument(
        '--cfg-options',
        nargs='+',
        action=DictAction,
        help='override some settings in the used config, the key-value pair '
        'in xxx=yyy format will be merged into config file. If the value to '
        'be overwritten is a list, it should be like key="[a,b]" or key=a,b '
        'It also allows nested list/tuple values, e.g. key="[(a,b),(c,d)]" '
        'Note that the quotation marks are necessary and that no white space '
        'is allowed.')
    args = parser.parse_args()
    return args


def main():
    args = parse_args()
    cfg = Config.fromfile(args.config)
    if args.cfg_options is not None:
        cfg.merge_from_dict(args.cfg_options)

    # register all modules in mmtrack into the registries
    register_all_modules(init_default_scope=True)

    dataset = DATASETS.build(cfg.train_dataloader.dataset)

    visualizer = VISUALIZERS.build(cfg.visualizer)
    visualizer.dataset_meta = dataset.metainfo

    progress_bar = mmengine.ProgressBar(len(dataset))
    gt_sample = TrackDataSample()  # just to wrap the `gt_instances`
    for idx, item in enumerate(dataset):
        data_sample = item['data_samples']
        for img_key, imgs in item['inputs'].items():
            img_paths = data_sample.get(img_key + '_path')
            img_key_prefix = img_key[:-3]
            gt_instances = data_sample.get(img_key_prefix + 'gt_instances')
            if not isinstance(img_paths, list):
                img_paths = [img_paths]
            for img_idx in range(imgs.shape[0]):
                new_gt_instances = InstanceData()
                img_path = img_paths[img_idx]
                img = imgs[img_idx].permute(1, 2, 0).numpy()
                # For each item, their file names may be the same.
                # Create a new folder to avoid overwriting the image files.
                out_file = osp.join(args.output_dir,
                                    str(idx).zfill(6),
                                    f'{img_key_prefix}img_{img_idx}.jpg'
                                    ) if args.output_dir is not None else None

                img = img[..., [2, 1, 0]]  # bgr to rgb
                # Get the correct index for each instance by using
                # map_instances_to_img_idx
                map_instances_to_img_idx = gt_instances.\
                    map_instances_to_img_idx.numpy()
                idx_bool_flag = (map_instances_to_img_idx == img_idx)
                for key in ['bboxes', 'labels', 'instances_id']:
                    if key in gt_instances:
                        new_gt_instances[key] = gt_instances[key][
                            idx_bool_flag]

                gt_masks = gt_instances.get('masks', None)
                if gt_masks is not None:
                    gt_masks = gt_masks[idx_bool_flag]
                    masks = mask2ndarray(gt_masks)
                    new_gt_instances['masks'] = masks.astype(np.bool)

                gt_sample.gt_instances = new_gt_instances

                visualizer.add_datasample(
                    osp.basename(img_path),
                    img,
                    data_sample=gt_sample,
                    draw_pred=False,
                    show=not args.not_show,
                    wait_time=args.show_interval,
                    out_file=out_file)
                # Record file path mapping.
                if args.output_dir is not None:
                    with open(
                            osp.join(args.output_dir,
                                     str(idx).zfill(6), 'info.txt'), 'a') as f:
                        f.write(f'The source filepath of'
                                f' `{img_key_prefix}img_{img_idx}.jpg`'
                                f' is `{img_path}`.\n')

        progress_bar.update()


if __name__ == '__main__':
    main()



# Copyright (c) OpenMMLab. All rights reserved.
from typing import Optional

from mmtrack.registry import MODELS, TASK_UTILS
from mmtrack.utils import OptConfigType
from .deep_sort import DeepSORT


@MODELS.register_module()
class StrongSORT(DeepSORT):
    """StrongSORT: Make DeepSORT Great Again.

    Details can be found at `StrongSORT<https://arxiv.org/abs/2202.13514>`_.

    Args:
        detector (dict): Configuration of detector. Defaults to None.
        reid (dict): Configuration of reid. Defaults to None
        tracker (dict): Configuration of tracker. Defaults to None.
        kalman (dict): Configuration of Kalman filter. Defaults to None.
        cmc (dict): Configuration of camera model compensation.
            Defaults to None.
        data_preprocessor (dict or ConfigDict, optional): The pre-process
           config of :class:`TrackDataPreprocessor`.  it usually includes,
            ``pad_size_divisor``, ``pad_value``, ``mean`` and ``std``.
        init_cfg (dict or list[dict]): Configuration of initialization.
            Defaults to None.
    """

    def __init__(self,
                 detector: Optional[dict] = None,
                 reid: Optional[dict] = None,
                 tracker: Optional[dict] = None,
                 kalman: Optional[dict] = None,
                 cmc: Optional[dict] = None,
                 data_preprocessor: OptConfigType = None,
                 init_cfg: OptConfigType = None):
        super().__init__(detector, reid, tracker, kalman, data_preprocessor,
                         init_cfg)

        if kalman is not None:
            self.kalman = TASK_UTILS.build(kalman)

        if cmc is not None:
            self.cmc = TASK_UTILS.build(cmc)

    @property
    def with_cmc(self):
        """bool: whether the framework has a camera model compensation
                model.
        """
        return hasattr(self, 'cmc') and self.cmc is not None

    @property
    def with_kalman_filter(self):
        """bool: whether the framework has a Kalman filter."""
        return hasattr(self, 'kalman') and self.kalman is not None



_base_ = [
    '../../_base_/datasets/got10k.py',
    './stark-st1_r50_8xb16-500e_got10k-lasot-trackingnet-coco_base.py'
]

train_pipeline = {{_base_.train_pipeline}}

# dataset settings
data_root = {{_base_.data_root}}
train_dataloader = dict(
    dataset=dict(
        _delete_=True,
        type='GOT10kDataset',
        data_root=data_root,
        ann_file='GOT10k/annotations/got10k_train_infos.txt',
        data_prefix=dict(img_path='GOT10k'),
        pipeline=train_pipeline,
        test_mode=False))

# evaluator
val_evaluator = dict(outfile_prefix='results/stark_st1_got10k')
test_evaluator = val_evaluator



# Copyright (c) OpenMMLab. All rights reserved.
from .benchmark import (DataLoaderBenchmark, DatasetBenchmark,
                        InferenceBenchmark)
from .collect_env import collect_env
from .image import crop_image, gauss_blur, imrenormalize
from .misc import (convert_data_sample_type, format_video_level_show,
                   max_last2d, stack_batch)
from .mot_error_visualization import imshow_mot_errors
from .plot_sot_curve import (plot_norm_precision_curve, plot_precision_curve,
                             plot_success_curve)
from .setup_env import register_all_modules
from .typing import (ConfigType, ForwardResults, InstanceList, MultiConfig,
                     OptConfigType, OptInstanceList, OptMultiConfig,
                     OptSampleList, SampleList)

__all__ = [
    'collect_env', 'register_all_modules', 'DataLoaderBenchmark',
    'DatasetBenchmark', 'InferenceBenchmark', 'crop_image', 'imrenormalize',
    'stack_batch', 'ConfigType', 'ForwardResults', 'InstanceList',
    'MultiConfig', 'OptConfigType', 'OptInstanceList', 'OptMultiConfig',
    'OptSampleList', 'SampleList', 'convert_data_sample_type',
    'imshow_mot_errors', 'max_last2d', 'gauss_blur', 'format_video_level_show',
    'plot_success_curve', 'plot_norm_precision_curve', 'plot_precision_curve'
]



# Copyright (c) OpenMMLab. All rights reserved.
from typing import Sequence, Union

import cv2
import mmcv
import numpy as np
import torch
import torch.nn.functional as F
from torch import Tensor


def crop_image(image, crop_region, crop_size, padding=(0, 0, 0)):
    """Crop image based on `crop_region` and `crop_size`.

    Args:
        image (ndarray): of shape (H, W, 3).
        crop_region (ndarray): of shape (4, ) in [x1, y1, x2, y2] format.
        crop_size (int): Crop size.
        padding (tuple | ndarray): of shape (3, ) denoting the padding values.

    Returns:
        ndarray: Cropped image of shape (crop_size, crop_size, 3).
    """
    a = crop_size / (crop_region[2] - crop_region[0])
    b = crop_size / (crop_region[3] - crop_region[1])
    c = -a * crop_region[0]
    d = -b * crop_region[1]
    mapping = np.array([[a, 0, c], [0, b, d]]).astype(np.float32)
    crop_image = cv2.warpAffine(
        image,
        mapping, (crop_size, crop_size),
        borderMode=cv2.BORDER_CONSTANT,
        borderValue=padding)
    return crop_image


def imrenormalize(img: Union[Tensor, np.ndarray], img_norm_cfg: dict,
                  new_img_norm_cfg: dict) -> Union[Tensor, np.ndarray]:
    """Re-normalize the image.

    Args:
        img (Tensor | ndarray): Input image. If the input is a Tensor, the
            shape is (1, C, H, W). If the input is a ndarray, the shape
            is (H, W, C).
        img_norm_cfg (dict): Original configuration for the normalization.
        new_img_norm_cfg (dict): New configuration for the normalization.

    Returns:
        Tensor | ndarray: Output image with the same type and shape of
        the input.
    """
    if isinstance(img, torch.Tensor):
        assert img.ndim == 4 and img.shape[0] == 1
        new_img = img.squeeze(0).cpu().numpy().transpose(1, 2, 0)
        new_img = _imrenormalize(new_img, img_norm_cfg, new_img_norm_cfg)
        new_img = new_img.transpose(2, 0, 1)[None]
        return torch.from_numpy(new_img).to(img)
    else:
        return _imrenormalize(img, img_norm_cfg, new_img_norm_cfg)


def _imrenormalize(img: Union[Tensor, np.ndarray], img_norm_cfg: dict,
                   new_img_norm_cfg: dict) -> Union[Tensor, np.ndarray]:
    """Re-normalize the image."""
    img_norm_cfg = img_norm_cfg.copy()
    new_img_norm_cfg = new_img_norm_cfg.copy()
    for k, v in img_norm_cfg.items():
        if (k == 'mean' or k == 'std') and not isinstance(v, np.ndarray):
            img_norm_cfg[k] = np.array(v, dtype=img.dtype)
    # reverse cfg
    if 'bgr_to_rgb' in img_norm_cfg:
        img_norm_cfg['rgb_to_bgr'] = img_norm_cfg['bgr_to_rgb']
        img_norm_cfg.pop('bgr_to_rgb')
    for k, v in new_img_norm_cfg.items():
        if (k == 'mean' or k == 'std') and not isinstance(v, np.ndarray):
            new_img_norm_cfg[k] = np.array(v, dtype=img.dtype)
    img = mmcv.imdenormalize(img, **img_norm_cfg)
    img = mmcv.imnormalize(img, **new_img_norm_cfg)
    return img


def gauss_blur(image: Tensor, kernel_size: Sequence,
               sigma: Sequence) -> Tensor:
    """The gauss blur transform.

    Args:
        image (Tensor): of shape (n, c, h, w)
        kernel_size (Tensor): The argument kernel size for gauss blur.
        sigma (Sequence): The argument sigma for gauss blur.

    Returns:
        Tensor: The blurred image.
    """
    assert len(kernel_size) == len(sigma) == 2
    x_coord = [
        torch.arange(-size, size + 1, dtype=torch.float32)
        for size in kernel_size
    ]
    filter = [
        torch.exp(-(x**2) / (2 * s**2)).to(image.device)
        for x, s in zip(x_coord, sigma)
    ]
    filter[0] = filter[0].view(1, 1, -1, 1) / filter[0].sum()
    filter[1] = filter[1].view(1, 1, 1, -1) / filter[1].sum()

    size = image.shape[2:]
    img_1 = F.conv2d(
        image.view(-1, 1, size[0], size[1]),
        filter[0],
        padding=(kernel_size[0], 0))

    img_2 = F.conv2d(
        img_1, filter[1],
        padding=(0, kernel_size[1])).view(1, -1, size[0], size[1])

    return img_2



# Copyright (c) OpenMMLab. All rights reserved.
import math
from typing import Dict, List, Optional, Tuple, Union

import cv2
import numpy as np
from mmcv.transforms import BaseTransform
from mmcv.transforms.utils import cache_randomness
from mmengine.logging import print_log

from mmtrack.registry import TRANSFORMS
from mmtrack.utils import crop_image


@TRANSFORMS.register_module()
class CropLikeSiamFC(BaseTransform):
    """Crop images as SiamFC did.

    The way of cropping an image is proposed in
    "Fully-Convolutional Siamese Networks for Object Tracking."
    `SiamFC <https://arxiv.org/abs/1606.09549>`_.

    Required Keys:

    - gt_bboxes (np.float32)
    - gt_bboxes_labels (np.int32)
    - gt_instances_id (np.int32)
    - gt_masks (BitmapMasks | PolygonMasks)
    - gt_seg_map (np.uint8)
    - gt_ignore_flags (np.bool)
    - img
    - img_shape (optional)

    Modified Keys:

    - gt_bboxes
    - img
    - img_shape (optional)

    Args:
        context_amount (float): The context amount around a bounding box.
            Defaults to 0.5.
        exemplar_size (int): Exemplar size. Defaults to 127.
        crop_size (int): Crop size. Defaults to 511.
    """

    def __init__(self,
                 context_amount: float = 0.5,
                 exemplar_size: int = 127,
                 crop_size: int = 511):
        self.context_amount = context_amount
        self.exemplar_size = exemplar_size
        self.crop_size = crop_size

    def crop_like_SiamFC(self,
                         image: np.ndarray,
                         bbox: np.ndarray,
                         context_amount: float = 0.5,
                         exemplar_size: int = 127,
                         crop_size: int = 511) -> np.ndarray:
        """Crop an image as SiamFC did.

        Args:
            image (np.ndarray): of shape (H, W, 3).
            bbox (np.ndarray): of shape (4, ) in [x1, y1, x2, y2] format.
            context_amount (float): The context amount around a bounding box.
                Defaults to 0.5.
            exemplar_size (int): Exemplar size. Defaults to 127.
            crop_size (int): Crop size. Defaults to 511.

        Returns:
            np.ndarray: The cropped image of shape (crop_size, crop_size, 3).
        """
        padding = np.mean(image, axis=(0, 1)).tolist()

        bbox = np.array([
            0.5 * (bbox[2] + bbox[0]), 0.5 * (bbox[3] + bbox[1]),
            bbox[2] - bbox[0], bbox[3] - bbox[1]
        ])
        z_width = bbox[2] + context_amount * (bbox[2] + bbox[3])
        z_height = bbox[3] + context_amount * (bbox[2] + bbox[3])
        z_size = np.sqrt(z_width * z_height)

        z_scale = exemplar_size / z_size
        d_search = (crop_size - exemplar_size) / 2.
        pad = d_search / z_scale
        x_size = z_size + 2 * pad
        x_bbox = np.array([
            bbox[0] - 0.5 * x_size, bbox[1] - 0.5 * x_size,
            bbox[0] + 0.5 * x_size, bbox[1] + 0.5 * x_size
        ])

        x_crop_img = crop_image(image, x_bbox, crop_size, padding)
        return x_crop_img

    def generate_box(self, image: np.ndarray, gt_bbox: np.ndarray,
                     context_amount: float, exemplar_size: int) -> np.ndarray:
        """Generate box based on cropped image.

        Args:
            image (np.ndarray): The cropped image of shape
                (self.crop_size, self.crop_size, 3).
            gt_bbox (np.ndarray): In shape (4, ), in [x1, y1, x2, y2] format.
            context_amount (float): The context amount around a bounding box.
            exemplar_size (int): Exemplar size. Defaults to 127.

        Returns:
            np.ndarray: Generated box of shape (4, ) in [x1, y1, x2, y2]
                format.
        """
        img_h, img_w = image.shape[:2]
        w, h = gt_bbox[2] - gt_bbox[0], gt_bbox[3] - gt_bbox[1]

        z_width = w + context_amount * (w + h)
        z_height = h + context_amount * (w + h)
        z_scale = np.sqrt(z_width * z_height)
        z_scale_factor = exemplar_size / z_scale
        w = w * z_scale_factor
        h = h * z_scale_factor
        cx, cy = img_w // 2, img_h // 2
        bbox = np.array(
            [cx - 0.5 * w, cy - 0.5 * h, cx + 0.5 * w, cy + 0.5 * h],
            dtype=np.float32)

        return bbox

    def transform(self, results: dict) -> dict:
        """The transform function.

        For each dict in results, crop image like SiamFC did.

        Args:
            results (dict): Dict from :obj:`mmtrack.dataset.BaseSOTDataset`.

        Returns:
            dict: Dict that contains cropped images and
                corresponding ground truth boxes.
        """
        crop_img = self.crop_like_SiamFC(results['img'],
                                         results['gt_bboxes'].squeeze(),
                                         self.context_amount,
                                         self.exemplar_size, self.crop_size)
        generated_bbox = self.generate_box(crop_img,
                                           results['gt_bboxes'].squeeze(),
                                           self.context_amount,
                                           self.exemplar_size)
        if 'img_shape' in results:
            results['img_shape'] = crop_img.shape

        results['img'] = crop_img
        results['gt_bboxes'] = generated_bbox[None]

        return results

    def __repr__(self) -> str:
        repr_str = self.__class__.__name__
        repr_str += f'(context_amount={self.context_amount}, '
        repr_str += f'exemplar_size={self.exemplar_size}, '
        repr_str += f'crop_size={self.crop_size})'
        return repr_str


@TRANSFORMS.register_module()
class SeqCropLikeStark(BaseTransform):
    """Crop images as Stark did.

    The way of cropping an image is proposed in
    "Learning Spatio-Temporal Transformer for Visual Tracking."
    `Stark <https://arxiv.org/abs/2103.17154>`_.

    Required Keys:

    - gt_bboxes (np.float32)
    - gt_bboxes_labels (np.int32)
    - gt_instances_id (np.int32)
    - gt_masks (BitmapMasks | PolygonMasks)
    - gt_seg_map (np.uint8)
    - gt_ignore_flags (np.bool)
    - img
    - img_shape (optional)
    - jittered_bboxes (np.float32)

    Modified Keys:

    - gt_bboxes
    - img
    - img_shape (optional)

    Added keys:

    - padding_mask

    Args:
        crop_size_factor (list[int | float]): contains the ratio of crop size
            to bbox size.
        output_size (list[int | float]): contains the size of resized image
            (always square).
    """

    def __init__(self, crop_size_factor: List[Union[int, float]],
                 output_size: List[Union[int, float]]):
        self.crop_size_factor = crop_size_factor
        self.output_size = output_size

    def crop_like_stark(
            self, img: np.ndarray, bbox: np.ndarray,
            crop_size_factor: np.ndarray,
            output_size: int) -> Union[np.ndarray, float, np.ndarray]:
        """Crop an image as Stark did.

        Args:
            img (np.ndarray): of shape (H, W, 3).
            bbox (np.ndarray): of shape (4, ) in [x1, y1, x2, y2] format.
            crop_size_factor (float): the ratio of crop size to bbox size
            output_size (int): the size of resized image (always square).

        Returns:
            img_crop_padded (np.ndarray): the cropped image of shape
                (crop_size, crop_size, 3).
            resize_factor (float): the ratio of original image scale to cropped
                image scale.
            pdding_mask (np.ndarray): the padding mask caused by cropping.
        """
        x1, y1, x2, y2 = np.split(bbox, 4, axis=-1)
        bbox_w, bbox_h = x2 - x1, y2 - y1
        cx, cy = x1 + bbox_w / 2., y1 + bbox_h / 2.

        img_h, img_w, _ = img.shape
        # 1. Crop image
        # 1.1 calculate crop size and pad size
        crop_size = math.ceil(math.sqrt(bbox_w * bbox_h) * crop_size_factor)
        crop_size = max(crop_size, 1)

        x1 = int(np.round(cx - crop_size * 0.5))
        x2 = x1 + crop_size
        y1 = int(np.round(cy - crop_size * 0.5))
        y2 = y1 + crop_size

        x1_pad = max(0, -x1)
        x2_pad = max(x2 - img_w + 1, 0)
        y1_pad = max(0, -y1)
        y2_pad = max(y2 - img_h + 1, 0)

        # 1.2 crop image
        img_crop = img[y1 + y1_pad:y2 - y2_pad, x1 + x1_pad:x2 - x2_pad, :]

        # 1.3 pad image
        img_crop_padded = cv2.copyMakeBorder(img_crop, y1_pad, y2_pad, x1_pad,
                                             x2_pad, cv2.BORDER_CONSTANT)
        # 1.4 generate padding mask
        img_h, img_w, _ = img_crop_padded.shape
        pdding_mask = np.ones((img_h, img_w))
        end_x, end_y = -x2_pad, -y2_pad
        if y2_pad == 0:
            end_y = None
        if x2_pad == 0:
            end_x = None
        pdding_mask[y1_pad:end_y, x1_pad:end_x] = 0

        # 2. Resize image and padding mask
        resize_factor = output_size / crop_size
        img_crop_padded = cv2.resize(img_crop_padded,
                                     (output_size, output_size))
        pdding_mask = cv2.resize(pdding_mask,
                                 (output_size, output_size)).astype(np.bool_)

        return img_crop_padded, resize_factor, pdding_mask

    def generate_box(self,
                     bbox_gt: np.ndarray,
                     bbox_cropped: np.ndarray,
                     resize_factor: float,
                     output_size: float,
                     normalize: bool = False) -> np.ndarray:
        """Transform the box coordinates from the original image coordinates to
        the coordinates of the cropped image.

        Args:
            bbox_gt (np.ndarray): of shape (4, ) in [x1, y1, x2, y2] format.
            bbox_cropped (np.ndarray): of shape (4, ) in [x1, y1, x2, y2]
                format.
            resize_factor (float): the ratio of original image scale to cropped
                image scale.
            output_size (float): the size of output image.
            normalize (bool): whether to normalize the output box.
                Defaults to False.

        Returns:
            np.ndarray: generated box of shape (4, ) in [x1, y1, x2, y2]
                format.
        """
        assert output_size > 0
        bbox_gt_center = (bbox_gt[0:2] + bbox_gt[2:4]) * 0.5
        bbox_cropped_center = (bbox_cropped[0:2] + bbox_cropped[2:4]) * 0.5

        bbox_out_center = (output_size - 1) / 2. + (
            bbox_gt_center - bbox_cropped_center) * resize_factor
        bbox_out_wh = (bbox_gt[2:4] - bbox_gt[0:2]) * resize_factor
        bbox_out = np.concatenate((bbox_out_center - 0.5 * bbox_out_wh,
                                   bbox_out_center + 0.5 * bbox_out_wh),
                                  axis=-1)

        return bbox_out / output_size if normalize else bbox_out

    def transform(self, results: dict) -> dict:
        """The transform function. For each dict in results, crop image like
        Stark did.

        Args:
            results (dict): Dict of list from
                :obj:`mmtrack.datasets.SeqBboxJitter`.

        Returns:
            dict: Dict of list that contains cropped image and
                the corresponding groundtruth bbox.
        """
        imgs = results['img']
        gt_bboxes = results['gt_bboxes']
        jittered_bboxes = results['jittered_bboxes']
        new_imgs = []
        results['padding_mask'] = []
        for i, (img, gt_bbox, jittered_bbox) in enumerate(
                zip(imgs, gt_bboxes, jittered_bboxes)):
            gt_bbox, jittered_bbox = gt_bbox[0], jittered_bbox[0]
            crop_img, resize_factor, padding_mask = self.crop_like_stark(
                img, jittered_bbox, self.crop_size_factor[i],
                self.output_size[i])

            generated_bbox = self.generate_box(
                gt_bbox,
                jittered_bbox,
                resize_factor,
                self.output_size[i],
                normalize=False)

            new_imgs.append(crop_img)
            if 'img_shape' in results:
                results['img_shape'][i] = crop_img.shape
            results['gt_bboxes'][i] = generated_bbox[None]
            results['padding_mask'].append(padding_mask)

        results['img'] = new_imgs

        return results

    def __repr__(self) -> str:
        repr_str = self.__class__.__name__
        repr_str += f'crop_size_factor={self.crop_size_factor}, '
        repr_str += f'output_size={self.output_size})'
        return repr_str


@TRANSFORMS.register_module()
class CropLikeDiMP(BaseTransform):
    """Crop images as PrDiMP did.

    The way of cropping an image is proposed in
    "Learning Discriminative Model Prediction for Tracking."
    `DiMP <https://arxiv.org/abs/1904.07220>`_.

    Args:
        crop_size_factor (float): contains the ratio of crop size
            to bbox size.
        output_size (float): contains the size of resized image
            (always square).
    """

    def __init__(self, crop_size_factor: float, output_size: float):
        self.crop_size_factor = crop_size_factor
        self.output_size = output_size

    def crop_like_dimp(
            self, img: np.ndarray, bbox: np.ndarray, crop_size_factor: float,
            output_size: int) -> Tuple[np.ndarray, np.ndarray, float]:
        """Crop an image as DiMP did.

        Note: The difference between dimp and stark is the operation of moving
        box inside image in dimp. This may cause the cropped image is not
        centered on the `bbox`.

        Args:
            image (np.ndarray): of shape (H, W, 3).
            bbox (np.ndarray): of shape (4, ) in [x1, y1, x2, y2] format.
            crop_size_factor (float): the ratio of crop size to bbox size
            output_size (int): the size of resized image (always square).

        Returns:
            img_crop_padded (np.ndarray): the cropped image of shape
                (crop_size, crop_size, 3).
            resize_factor (float): the ratio of original image scale to cropped
                image scale.
            pdding_mask (np.ndarray): the padding mask caused by cropping.
        """
        x1, y1, x2, y2 = np.split(bbox, 4, axis=-1)
        bbox_w, bbox_h = x2 - x1, y2 - y1
        cx, cy = x1 + bbox_w / 2., y1 + bbox_h / 2.

        img_h, img_w, _ = img.shape
        # 1. Crop image
        # 1.1 calculate crop size
        crop_size = math.ceil(math.sqrt(bbox_w * bbox_h) * crop_size_factor)
        crop_size = max(crop_size, 1)

        x1 = int(np.round(cx - crop_size * 0.5))
        x2 = x1 + crop_size
        y1 = int(np.round(cy - crop_size * 0.5))
        y2 = y1 + crop_size

        # 1.2 Move box inside image
        shift_x = max(0, -x1) + min(0, img_w - x2)
        x1 += shift_x
        x2 += shift_x

        shift_y = max(0, -y1) + min(0, img_h - y2)
        y1 += shift_y
        y2 += shift_y

        # keep the balance of left and right spacing if crop area exceeds the
        # image
        out_x = (max(0, -x1) + max(0, x2 - img_w)) // 2
        out_y = (max(0, -y1) + max(0, y2 - img_h)) // 2
        shift_x = (-x1 - out_x) * (out_x > 0)
        shift_y = (-y1 - out_y) * (out_y > 0)

        x1 += shift_x
        x2 += shift_x
        y1 += shift_y
        y2 += shift_y

        # 1.3 pad size
        x1_pad = max(0, -x1)
        x2_pad = max(x2 - img_w + 1, 0)
        y1_pad = max(0, -y1)
        y2_pad = max(y2 - img_h + 1, 0)

        # 1.4 crop image
        img_crop = img[y1 + y1_pad:y2 - y2_pad, x1 + x1_pad:x2 - x2_pad, :]

        # 1.5 pad image
        img_crop_padded = cv2.copyMakeBorder(img_crop, y1_pad, y2_pad, x1_pad,
                                             x2_pad, cv2.BORDER_REPLICATE)

        # 2. Resize image and padding mask
        assert y2 - y1 == crop_size
        assert x2 - x1 == crop_size
        resize_factor = output_size / crop_size
        img_crop_padded = cv2.resize(img_crop_padded,
                                     (output_size, output_size))

        # the new box of cropped area
        crop_area_bbox = np.array([x1, y1, x2 - x1, y2 - y1], dtype=float)

        return img_crop_padded, crop_area_bbox, resize_factor

    def generate_box(self, bbox_gt: np.ndarray, crop_area_bbox: np.ndarray,
                     resize_factor: np.ndarray) -> np.ndarray:
        """Transform the box coordinates from the original image coordinates to
        the coordinates of the resized cropped image. The center of cropped
        image may be not jittered bbox since the operation of moving box inside
        image.

        Args:
            bbox_gt (np.ndarray): of shape (4, ) in [x1, y1, x2, y2] format.
            crop_area_bbox (np.ndarray): of shape (4, ) in [x1, y1, w, h]
                format.
            resize_factor (float): the ratio of original image scale to cropped
                image scale.
            output_size (float): the size of output image.
            normalize (bool): whether to normalize the output box.
                Default to True.

        Returns:
            np.ndarray: generated box of shape (4, ) in [x1, y1, x2, y2]
                format.
        """
        bbox_out = bbox_gt.copy()
        # The coordinate origin of `bbox_out` is the top left corner of
        # `crop_area_bbox`.
        bbox_out[0:4:2] -= crop_area_bbox[0]
        bbox_out[1:4:2] -= crop_area_bbox[1]

        bbox_out *= resize_factor

        return bbox_out

    def transform(self, results: dict) -> dict:
        """Call function. Crop image like DiMP did.

        Args:
            results (dict): Dict from :obj:`mmtrack.dataset.BaseSOTDataset`.

        Returns:
            dict: Dict that contains cropped images and
                corresponding ground truth boxes.
        """
        gt_bbox = results['gt_bboxes'][0]
        jittered_bboxes = results['jittered_bboxes'][0]
        crop_img, crop_area_bbox, resize_factor = self.crop_like_dimp(
            results['img'], jittered_bboxes, self.crop_size_factor,
            self.output_size)

        generated_bbox = self.generate_box(gt_bbox, crop_area_bbox,
                                           resize_factor)

        results['img'] = crop_img
        if 'img_shape' in results:
            results['img_shape'] = crop_img.shape
        results['gt_bboxes'] = generated_bbox[None]

        return results


@TRANSFORMS.register_module()
class SeqBboxJitter(BaseTransform):
    """Bounding box jitter augmentation. The jittered bboxes are used for
    subsequent image cropping, like `SeqCropLikeStark`.

    Required Keys:

    - gt_bboxes
    - gt_bboxes_labels (optional)
    - gt_instances_id (optional)
    - gt_masks (optional)
    - gt_seg_map (optional)
    - gt_ignore_flags (optional)
    - img
    - img_shape (optional)

    Added Keys:

    - jittered_bboxes

    Args:
        scale_jitter_factor (list[int | float]): contains the factor of scale
            jitter.
        center_jitter_factor (list[int | float]): contains the factor of center
            jitter.
        crop_size_factor (list[int | float]): contains the ratio of crop size
            to bbox size.
    """

    def __init__(self, scale_jitter_factor: List[Union[int, float]],
                 center_jitter_factor: List[Union[int, float]],
                 crop_size_factor: List[Union[int, float]]):
        self.scale_jitter_factor = scale_jitter_factor
        self.center_jitter_factor = center_jitter_factor
        self.crop_size_factor = crop_size_factor

    def transform(self, results: Dict[str, List]) -> Optional[dict]:
        """The transform function.

        Args:
            results (Dict[str, List]): Dict of list from
                :obj:`mmtrack.datasets.BaseSOTDataset`.

        Returns:
            Optional[dict]: Dict of list that contains augmented images. If
                getting invalid cropped image, return None.
        """
        gt_bboxes = results['gt_bboxes']
        jittered_bboxes = []
        for i, gt_bbox in enumerate(gt_bboxes):
            x1, y1, x2, y2 = np.split(gt_bbox.squeeze(), 4, axis=-1)
            bbox_w, bbox_h = x2 - x1, y2 - y1
            gt_bbox_cxcywh = np.concatenate(
                [x1 + bbox_w / 2., y1 + bbox_h / 2., bbox_w, bbox_h], axis=-1)

            crop_img_size = -1
            # avoid croped image size too small.
            count = 0
            while crop_img_size < 1:
                count += 1
                if count > 100:
                    print_log(
                        f'-------- bbox {gt_bbox_cxcywh} is invalid -------')
                    return None
                jittered_wh = gt_bbox_cxcywh[2:4] * np.exp(
                    np.random.randn(2) * self.scale_jitter_factor[i])
                crop_img_size = np.ceil(
                    np.sqrt(jittered_wh.prod()) * self.crop_size_factor[i])

            max_offset = np.sqrt(
                jittered_wh.prod()) * self.center_jitter_factor[i]
            jittered_center = gt_bbox_cxcywh[0:2] + max_offset * (
                np.random.rand(2) - 0.5)

            jittered_bbox = np.concatenate(
                (jittered_center - 0.5 * jittered_wh,
                 jittered_center + 0.5 * jittered_wh),
                axis=-1)
            jittered_bboxes.append(jittered_bbox[None])

        results['jittered_bboxes'] = jittered_bboxes
        return results

    def __repr__(self) -> str:
        repr_str = self.__class__.__name__
        repr_str += f'scale_jitter_factor={self.scale_jitter_factor}, '
        repr_str += f'center_jitter_factor={self.center_jitter_factor}, '
        repr_str += f'crop_size_factor={self.crop_size_factor})'
        return repr_str


@TRANSFORMS.register_module()
class BrightnessAug(BaseTransform):
    """Brightness augmention for images.

    Required Keys:

    - gt_bboxes
    - gt_bboxes_labels (optional)
    - gt_instances_id (optional)
    - gt_masks (optional)
    - gt_seg_map (optional)
    - gt_ignore_flags (optional)
    - img
    - img_shape (optional)

    Modified Keys:

    - img

    Args:
        jitter_range (float): The range of brightness jitter.
            Defaults to 0..
    """

    def __init__(self, jitter_range: float = 0.):
        self.jitter_range = jitter_range

    @cache_randomness
    def _random_brightness_factor(self) -> float:
        """Generate the factor of brightness randomly.

        Returns:
            float: The factor of brightness.
        """

        brightness_factor = np.random.uniform(
            max(0, 1 - self.jitter_range), 1 + self.jitter_range)
        return brightness_factor

    def transform(self, results: dict) -> dict:
        """The transform function.

        For each dict in results, perform brightness augmention for image in
        the dict.

        Args:
            results (dict): list of dict from :obj:`mmengine.BaseDataset`.
        Returns:
            dict: Dict that contains augmented image.
        """
        brightness_factor = self._random_brightness_factor()
        image = np.dot(results['img'], brightness_factor).clip(0, 255.0)
        results['img'] = image

        return results

    def __repr__(self) -> str:
        repr_str = self.__class__.__name__
        repr_str += f'jitter_range={self.jitter_range})'
        return repr_str


@TRANSFORMS.register_module()
class GrayAug(BaseTransform):
    """Gray augmention for images.

    Required Keys:

    - gt_bboxes
    - gt_bboxes_labels (optional)
    - gt_instances_id (optional)
    - gt_masks (optional)
    - gt_seg_map (optional)
    - gt_ignore_flags (optional)
    - img
    - img_shape (optional)

    Modified Keys:

    - img

    Args:
        prob (float): The probability to perform gray augmention.
            Defaults to 0..
    """

    def __init__(self, prob: float = 0.):
        self.prob = prob

    @cache_randomness
    def _random_gray(self) -> bool:
        """Whether to convert the original image to gray image.

        Returns:
            bool: Whether to convert the original image to gray image.
        """
        if self.prob > np.random.random():
            convert2gray = True
        else:
            convert2gray = False
        return convert2gray

    def transform(self, results: dict) -> dict:
        """The transform function.

        For each dict in results, perform gray augmention for image in the
        dict.

        Args:
            results (list[dict]): List of dict from
                :obj:`mmengine.BaseDataset`.

        Returns:
            dict: Dict that contains augmented gray image.
        """
        if self._random_gray():
            grayed = cv2.cvtColor(results['img'], cv2.COLOR_BGR2GRAY)
            image = cv2.cvtColor(grayed, cv2.COLOR_GRAY2BGR)
            results['img'] = image

        return results

    def __repr__(self) -> str:
        repr_str = self.__class__.__name__
        repr_str += f'prob={self.prob})'
        return repr_str


@TRANSFORMS.register_module()
class SeqShiftScaleAug(BaseTransform):
    """Shift and rescale images and bounding boxes.

    Required Keys:

    - gt_bboxes
    - gt_bboxes_labels (optional)
    - gt_instances_id (optional)
    - gt_masks (optional)
    - gt_seg_map (optional)
    - gt_ignore_flags (optional)
    - img
    - img_shape (optional)

    Modified Keys:

    - gt_bboxes
    - img
    - img_shape (optional)

    Args:
        target_size (list[int]): list of int denoting exemplar size and search
            size, respectively. Defaults to [127, 255].
        shift (list[int]): list of int denoting the max shift offset. Defaults
            to [4, 64].
        scale (list[float]): list of float denoting the max rescale factor.
            Defaults to [0.05, 0.18].
    """

    def __init__(self,
                 target_size: List[int] = [127, 255],
                 shift: List[int] = [4, 64],
                 scale: List[float] = [0.05, 0.18]):
        self.target_size = target_size
        self.shift = shift
        self.scale = scale

    def _shift_scale_aug(self, image: np.ndarray, bbox: np.ndarray,
                         target_size: int, shift: int,
                         scale: float) -> Tuple[np.ndarray, np.ndarray]:
        """Shift and rescale an image and corresponding bounding box.

        Args:
            image (np.ndarray): of shape (H, W, 3). Typically H and W equal to
                511.
            bbox (np.ndarray): of shape (4, ) in [x1, y1, x2, y2] format.
            target_size (int): Exemplar size or search size.
            shift (int): The max shift offset.
            scale (float): The max rescale factor.

        Returns:
            tuple(np.ndarray, np.ndarray): The first element is of shape
            (target_size, target_size, 3), and the second element is the
            corresponding ground truth box in [x1, y1, x2, y2] format.
        """
        img_h, img_w = image.shape[:2]

        scale_x = (2 * np.random.random() - 1) * scale + 1
        scale_y = (2 * np.random.random() - 1) * scale + 1
        scale_x = min(scale_x, float(img_w) / target_size)
        scale_y = min(scale_y, float(img_h) / target_size)
        crop_region = np.array([
            img_w // 2 - 0.5 * scale_x * target_size,
            img_h // 2 - 0.5 * scale_y * target_size,
            img_w // 2 + 0.5 * scale_x * target_size,
            img_h // 2 + 0.5 * scale_y * target_size
        ])

        shift_x = (2 * np.random.random() - 1) * shift
        shift_y = (2 * np.random.random() - 1) * shift
        shift_x = max(-crop_region[0], min(img_w - crop_region[2], shift_x))
        shift_y = max(-crop_region[1], min(img_h - crop_region[3], shift_y))
        shift = np.array([shift_x, shift_y, shift_x, shift_y])
        crop_region += shift

        crop_img = crop_image(image, crop_region, target_size)
        bbox -= np.array(
            [crop_region[0], crop_region[1], crop_region[0], crop_region[1]])
        bbox /= np.array([scale_x, scale_y, scale_x, scale_y],
                         dtype=np.float32)
        return crop_img, bbox

    def transform(self, results: dict) -> dict:
        """The transform function.

        For each dict in results, shift and rescale the image and the
        bounding box in the dict.

        Args:
            results (dict(list)): Dict of list that from
                :obj:`mmtrack.dataset.BaseSOTDataset`.

        Returns:
            dict(list): List of dict that contains cropped image and
            corresponding ground truth box.
        """
        imgs = results['img']
        gt_bboxes = results['gt_bboxes']
        new_imgs = []
        new_gt_bboxes = []
        for i, (img, gt_bbox) in enumerate(zip(imgs, gt_bboxes)):
            crop_img, crop_bbox = self._shift_scale_aug(
                img, gt_bbox.squeeze(), self.target_size[i], self.shift[i],
                self.scale[i])
            crop_bbox = crop_bbox[None]
            new_gt_bboxes.append(crop_bbox)
            new_imgs.append(crop_img)
            if 'img_shape' in results:
                results['img_shape'][i] = crop_img.shape
        results['img'] = new_imgs
        results['gt_bboxes'] = new_gt_bboxes
        return results

    def __repr__(self) -> str:
        repr_str = self.__class__.__name__
        repr_str += f'(target_size={self.target_size}, '
        repr_str += f'shift={self.shift}, '
        repr_str += f'scale={self.scale})'
        return repr_str


@TRANSFORMS.register_module()
class SeqColorAug(BaseTransform):
    """Color augmention for images.

    Required Keys:

    - gt_bboxes
    - gt_bboxes_labels (optional)
    - gt_instances_id (optional)
    - gt_masks (optional)
    - gt_seg_map (optional)
    - gt_ignore_flags (optional)
    - img
    - img_shape (optional)

    Modified Keys:

    - img

    Args:
        prob (list[float]): The probability to perform color augmention for
            each image. Defaults to [1.0, 1.0].
        rgb_var (list[list]]): The values of color augmentaion. Defaults to
            [[-0.55919361, 0.98062831, -0.41940627],
            [1.72091413, 0.19879334, -1.82968581],
            [4.64467907, 4.73710203, 4.88324118]].
    """

    def __init__(self,
                 prob: List[float] = [1.0, 1.0],
                 rgb_var: List[List] = [[-0.55919361, 0.98062831, -0.41940627],
                                        [1.72091413, 0.19879334, -1.82968581],
                                        [4.64467907, 4.73710203, 4.88324118]]):
        self.prob = prob
        self.rgb_var = np.array(rgb_var, dtype=np.float32)

    def transform(self, results: dict) -> dict:
        """The transform function.

        For each dict in results, perform color augmention for image in the
        dict.

        Args:
            results (dict[list]): Dict of list that from
                :obj:`mmengine.BaseDataset`.

        Returns:
            dict[list]: Dict of list that contains augmented color image.
        """
        imgs = results['img']
        new_imgs = []
        for i, img in enumerate(imgs):
            if self.prob[i] > np.random.random():
                offset = np.dot(self.rgb_var, np.random.randn(3, 1))
                # bgr to rgb
                offset = offset[::-1]
                offset = offset.reshape(3)
                img = (img - offset).astype(np.float32)
            new_imgs.append(img)

        results['img'] = new_imgs
        return results

    def __repr__(self) -> str:
        repr_str = self.__class__.__name__
        repr_str += f'(prob={self.prob}, '
        repr_str += f'rgb_var={self.rgb_var})'
        return repr_str


@TRANSFORMS.register_module()
class SeqBlurAug(BaseTransform):
    """Blur augmention for images.

    Required Keys:

    - gt_bboxes
    - gt_bboxes_labels (optional)
    - gt_instances_id (optional)
    - gt_masks (optional)
    - gt_seg_map (optional)
    - gt_ignore_flags (optional)
    - img
    - img_shape (optional)

    Modified Keys:

    - img

    Args:
        prob (list[float]): The probability to perform blur augmention for
            each image. Defaults to [0.0, 0.2].
    """

    def __init__(self, prob: List[float] = [0.0, 0.2]):
        self.prob = prob

    def transform(self, results: dict) -> dict:
        """The transform function.

        For each dict in results, perform blur augmention for image in the
        dict.

        Args:
            results (dict[list]): Dict of list that from
                :obj:`mmtrack.CocoVideoDataset`.

        Returns:
            dict[list]: Dict of list that contains augmented blur image.
        """
        imgs = results['img']
        new_imgs = []
        for i, img in enumerate(imgs):
            if self.prob[i] > np.random.random():
                sizes = np.arange(5, 46, 2)
                size = np.random.choice(sizes)
                kernel = np.zeros((size, size))
                c = int(size / 2)
                wx = np.random.random()
                kernel[:, c] += 1. / size * wx
                kernel[c, :] += 1. / size * (1 - wx)
                img = cv2.filter2D(img, -1, kernel)
            new_imgs.append(img)

        results['img'] = new_imgs
        return results

    def __repr__(self) -> str:
        repr_str = self.__class__.__name__
        repr_str += f'prob={self.prob})'
        return repr_str



# Copyright (c) OpenMMLab. All rights reserved.
from .eval_sot_ope import eval_sot_ope
from .eval_sot_vot import (bbox2region, eval_sot_accuracy_robustness,
                           eval_sot_eao)
from .ytvis import YTVIS
from .ytviseval import YTVISeval

__all__ = [
    'eval_sot_ope', 'bbox2region', 'eval_sot_eao',
    'eval_sot_accuracy_robustness', 'YTVIS', 'YTVISeval'
]



# Copyright (c) OpenMMLab. All rights reserved.
# The codes are modified from https://github.com/votchallenge/toolkit/blob/master/vot/analysis/supervised.py # noqa: E501
from typing import Dict, List, Optional, Tuple, Union

import numpy as np

try:
    import vot
    from vot.analysis import is_special
    from vot.region import Polygon, Rectangle, Special
    from vot.region import calculate_overlaps as calculate_region_overlaps
except ImportError:
    vot = None


def bbox2region(bbox: np.ndarray) -> 'Union[Rectangle, Polygon]':
    """Convert bbox to Rectangle or Polygon Class object.

    Args:
        bbox (np.ndarray): the format of rectangle bbox is
            [tl_x, tl_y, br_x, br_y];
            the format of polygon is [x1, y1, x2, y2, ...].

    Returns:
        :obj:`Rectangle` or :obj:`Polygon`.
    """
    if vot is None:
        raise ImportError(
            'Please run'
            ' pip install git+https://github.com/votchallenge/toolkit.git '
            'to manually install vot-toolkit')

    if len(bbox) == 1:
        return Special(bbox[0])
    elif len(bbox) == 4:
        return Rectangle(bbox[0], bbox[1], bbox[2] - bbox[0],
                         bbox[3] - bbox[1])
    elif len(bbox) % 2 == 0 and len(bbox) > 4:
        return Polygon([(x_, y_) for x_, y_ in zip(bbox[::2], bbox[1::2])])
    else:
        raise NotImplementedError(
            f'The length of bbox is {len(bbox)}, which is not supported')


def trajectory2region(trajectory: List[np.ndarray]) -> List:
    """Convert bbox trajectory to Rectangle or Polygon Class object trajectory.

    Args:
        trajectory (List[np.ndarray]): The outer list contains bbox of
            each frame in a video. The bbox is a ndarray.

    Returns:
        List: contains the Region Class object of each frame in a
            trajectory.
    """
    traj_region = []
    for bbox in trajectory:
        traj_region.append(bbox2region(bbox))
    return traj_region


def locate_failures_inits(trajectory: List[np.ndarray]) -> Tuple[List, List]:
    """locate the failure frame and initialized frame in a trajectory.

    Args:
        trajectory (List[np.ndarray]): list of tracking results.

    Returns:
        Tuple[List, List]:
            - fail_inds (List): index of failed frame in a trajectory.
            - init_inds (List): index of initialized frame in a trajectory.
    """
    fail_inds = []
    init_inds = []
    for i, bbox in enumerate(trajectory):
        if len(bbox) == 1:
            if bbox[0] == 1.:
                init_inds.append(i)
            elif bbox[0] == 2.:
                fail_inds.append(i)
    return fail_inds, init_inds


def count_failures(trajectory: List[np.ndarray]) -> List:
    """count the number of failed frame in a trajectory.

    Args:
        trajectory (List[np.ndarray]): list of tracking results.

    Returns:
        List: the number of failed frame in a trajectory.
    """
    num_fails = 0
    for bbox in trajectory:
        if len(bbox) == 1 and bbox[0] == 2.:
            num_fails += 1
    return num_fails


def calc_accuracy(gt_trajectory: List[List],
                  pred_trajectory: List[List],
                  burnin: int = 10,
                  ignore_unknown: bool = True,
                  video_wh: Optional[Tuple[int, int]] = None) -> float:
    """Calculate accuracy over the sequence.

    Args:
        gt_trajectory (List[List]): list of bboxes
        pred_trajectory (List[np.ndarray]): The outer list contains the
            tracking results of each frame in one video. The ndarray has two
            cases:
                - bbox: denotes the normal tracking box in
                    [tl_x, tl_y, br_x, br_y] format.
                - special tracking state: [0] denotes the unknown state,
                    namely the skipping frame after failure, [1] denotes the
                    initialized state, and [2] denotes the failed state.
        burnin (int, optional): number of frames that have to be ignored after
            the re-initialization when calculating accuracy. Default is 10.
        ignore_unknown (bool, optional): whether ignore the skipping frames
            after failures when calculating accuracy. Default is True.
        video_wh (Optional[Tuple[int, int]], optional): bounding region
            (width, height)

    Return:
        Float: accuracy over the sequence.
    """
    pred_traj_region = trajectory2region(pred_trajectory)
    gt_traj_region = trajectory2region(gt_trajectory)
    overlaps = np.array(
        calculate_region_overlaps(pred_traj_region, gt_traj_region, video_wh))
    mask = np.ones(len(overlaps), dtype=bool)

    for i, region in enumerate(pred_traj_region):
        if is_special(region, Special.UNKNOWN) and ignore_unknown:
            mask[i] = False
        elif is_special(region, Special.INITIALIZATION):
            for j in range(i, min(len(pred_traj_region), i + burnin)):
                mask[j] = False
        elif is_special(region, Special.FAILURE):
            mask[i] = False
    return np.mean(overlaps[mask]).item() if any(mask) else 0.


def eval_sot_accuracy_robustness(
        results: List[List[np.ndarray]],
        annotations: List[np.ndarray],
        burnin: int = 10,
        ignore_unknown: bool = True,
        videos_wh: Optional[Tuple[int, int]] = None) -> Dict[str, float]:
    """Calculate accuracy and robustness over all tracking sequences.

    Args:
        results (List[List[np.ndarray]]): The first list contains the
            tracking results of each video. The second list contains the
            tracking results of each frame in one video. The ndarray have two
            cases:
                - bbox: denotes the normal tracking box in
                    [tl_x, tl_y, br_x, br_y] format.
                - special tracking state: [0] denotes the unknown state,
                    namely the skipping frame after failure, [1] denotes the
                    initialized state, and [2] denotes the failed state.
        annotations (List[np.ndarray]): The list contains the gt_bboxes of each
            video. The ndarray is gt_bboxes of one video. It's in (N, 4) shape.
            Each bbox is in (x1, y1, w, h) format.
        burnin (int, optional): number of frames that have to be ignored after
            the re-initialization when calculating accuracy. Default is 10.
        ignore_unknown (bool, optional): whether ignore the skipping frames
            after failures when calculating accuracy. Default is True.
        videos_wh (Optional[Tuple[int, int]], optional): bounding region
            (width, height)

    Return:
        Dict[str: float]: accuracy and robustness in EAO evaluation metric.
    """
    if vot is None:
        raise ImportError(
            'Please run'
            'pip install git+https://github.com/votchallenge/toolkit.git'
            'to manually install vot-toolkit')

    accuracy = 0
    num_fails = 0
    weight = 0
    for i, (gt_traj, pred_traj) in enumerate(zip(annotations, results)):
        assert len(gt_traj) == len(pred_traj)
        assert len(pred_traj[0]) == 1 and pred_traj[0][0] == 1
        num_fails += count_failures(pred_traj)
        accuracy += calc_accuracy(
            gt_traj,
            pred_traj,
            burnin=burnin,
            ignore_unknown=ignore_unknown,
            video_wh=videos_wh[i]) * len(pred_traj)
        weight += len(pred_traj)

    accuracy /= weight
    robustness = num_fails / weight * 100
    return dict(accuracy=accuracy, robustness=robustness, num_fails=num_fails)


def calc_eao_curve(overlaps: List[List], successes: List) -> np.ndarray:
    """Calculate EAO curve over all tracking sequences.

    Args:
        overlaps (List[List]): The outer list contains the overlaps of each
            video. The inner list contains the overlap of each frame in one
            video.
        successes (List): The list contains the tracking states of last frame
            in each fragment.

    Return:
        np.ndarray: The N-th element in ndarray denotes the average overlaps
            from 1 to N in all fragments.
    """
    max_length = max([len(_) for _ in overlaps])
    total_runs = len(overlaps)

    overlaps_array = np.zeros((total_runs, max_length), dtype=np.float32)
    # mask out frames which are not considered in EAO calculation. initial
    # value are zero, meaning ignored.
    mask = np.zeros((total_runs, max_length), dtype=np.float32)
    for i, (overlap, success) in enumerate(zip(overlaps, successes)):
        overlaps_array[i, :len(overlap)] = np.array(overlap)
        if not success:
            # tracker has failed during this sequence - consider all of
            # 'overlaps_array' and use the default padding from the end of
            # sequence to max length.
            mask[i, :] = 1
        else:
            # tracker has successfully tracked to the end - consider only this
            # part of the true sequence, and ignore the padding from the end of
            # sequence to max length.
            mask[i, :len(overlap)] = 1

    overlaps_array_sum = overlaps_array.copy()
    # overlaps_array_sum[i,j] means the mean overlap from 1 to j in i-th
    # sequence
    for j in range(1, overlaps_array_sum.shape[1]):
        overlaps_array_sum[:, j] = np.mean(overlaps_array[:, 1:j + 1], axis=1)

    return np.sum(overlaps_array_sum * mask, axis=0) / np.sum(mask, axis=0)


def eval_sot_eao(
        results: List[List[np.ndarray]],
        annotations: List[np.ndarray],
        interval: Tuple[int, int] = [100, 356],
        videos_wh: Optional[Tuple[int, int]] = None) -> Dict[str, float]:
    """Calculate EAO socre over all tracking sequences.

    Args:
        results (List[List[np.ndarray]]): The first list contains the
            tracking results of each video. The second list contains the
            tracking results of each frame in one video. The ndarray have two
            cases:
                - bbox: denotes the normal tracking box in
                    [tl_x, tl_y, br_x, br_y] format.
                - special tracking state: [0] denotes the unknown state,
                    namely the skipping frame after failure, [1] denotes the
                    initialized state, and [2] denotes the failed state.
        annotations (list[np.ndarray]): The list contains the gt_bboxes of each
            video. The ndarray is gt_bboxes of one video. It's in (N, 4) shape.
            Each bbox is in (x1, y1, w, h) format.
        interval: an specified interval in EAO curve used to calculate the EAO
            score. There are different settings in different VOT challenge.
            Default is VOT2018 setting: [100, 356].
        videos_wh (list[tuple(width, height), ...]): The list contains the
            width and height of each video. Default is None.

    Return:
        Dict[str, float]: EAO score in EAO evaluation metric.
    """
    if vot is None:
        raise ImportError(
            'Please run'
            'pip install git+https://github.com/votchallenge/toolkit.git'
            'to manually install vot-toolkit')

    if videos_wh is None:
        videos_wh = [None] * len(annotations)

    all_overlaps = []
    all_successes = []

    for i, (gt_traj, pred_traj) in enumerate(zip(annotations, results)):
        assert len(gt_traj) == len(
            pred_traj), f'{len(gt_traj)} == {len(pred_traj)}'
        # initialized bbox annotation is [1]
        assert len(pred_traj[0]) == 1 and pred_traj[0][
            0] == 1, f'{len(pred_traj[0])} == 1 and {pred_traj[0][0]} == 1'
        fail_inds, init_inds = locate_failures_inits(pred_traj)
        pred_traj = trajectory2region(pred_traj)
        gt_traj = trajectory2region(gt_traj)
        overlaps = calculate_region_overlaps(pred_traj, gt_traj, videos_wh[i])

        if len(fail_inds) > 0:
            for i in range(len(fail_inds)):
                all_overlaps.append(overlaps[init_inds[i]:fail_inds[i]])
                all_successes.append(False)

            # handle last initialization
            if len(init_inds) > len(fail_inds):
                # tracker was initialized, but it has not failed until the end
                # of the sequence
                all_overlaps.append(overlaps[init_inds[-1]:])
                all_successes.append(True)
        else:
            all_overlaps.append(overlaps)
            all_successes.append(True)

    eao_curve = calc_eao_curve(all_overlaps, all_successes)
    eao_score = np.mean(eao_curve[interval[0]:interval[1] + 1]).item()
    eao = dict(eao=eao_score)
    return eao



# Copyright (c) OpenMMLab. All rights reserved.
from typing import List, Optional, Tuple, Union

import torch
from torch import Tensor
from torch.nn.modules.batchnorm import _BatchNorm
from torch.nn.modules.conv import _ConvNd

from mmtrack.registry import MODELS
from mmtrack.utils import (InstanceList, OptConfigType, OptMultiConfig,
                           SampleList)
from .base import BaseSingleObjectTracker


@MODELS.register_module()
class SiamRPN(BaseSingleObjectTracker):
    """SiamRPN++: Evolution of Siamese Visual Tracking with Very Deep Networks.

    This single object tracker is the implementation of `SiamRPN++
    <https://arxiv.org/abs/1812.11703>`_.
    """

    def __init__(self,
                 backbone: dict,
                 neck: Optional[dict] = None,
                 head: Optional[dict] = None,
                 train_cfg: Optional[dict] = None,
                 test_cfg: Optional[dict] = None,
                 frozen_modules: Optional[Union[List[str], Tuple[str],
                                                str]] = None,
                 data_preprocessor: OptConfigType = None,
                 init_cfg: OptMultiConfig = None):
        super(SiamRPN, self).__init__(data_preprocessor, init_cfg)
        self.backbone = MODELS.build(backbone)
        if neck is not None:
            self.neck = MODELS.build(neck)
        head = head.copy()
        head.update(train_cfg=train_cfg.rpn, test_cfg=test_cfg.rpn)
        self.head = MODELS.build(head)

        self.test_cfg = test_cfg
        self.train_cfg = train_cfg

        if frozen_modules is not None:
            self.freeze_module(frozen_modules)

    def init_weights(self):
        """Initialize the weights of modules in single object tracker."""
        # We don't use the `init_weights()` function in BaseModule, since it
        # doesn't support the initialization method from `reset_parameters()`
        # in Pytorch.
        if self.with_backbone:
            self.backbone.init_weights()

        if self.with_neck:
            for m in self.neck.modules():
                if isinstance(m, _ConvNd) or isinstance(m, _BatchNorm):
                    m.reset_parameters()

        if self.with_head:
            for m in self.head.modules():
                if isinstance(m, _ConvNd) or isinstance(m, _BatchNorm):
                    m.reset_parameters()

    def forward_template(self, z_img: Tensor) -> Tuple[Tensor]:
        """Extract the features of exemplar images.

        Args:
            z_img (Tensor): of shape (N, C, H, W) encoding input exemplar
                images. Typically H and W equal to 127.

        Returns:
            Tuple[Tensor, ...]: Multi level feature map of exemplar
                images.
        """
        z_feat = self.backbone(z_img)
        if self.with_neck:
            z_feat = self.neck(z_feat)

        z_feat_center = []
        for i in range(len(z_feat)):
            left = (z_feat[i].size(3) - self.test_cfg.center_size) // 2
            right = left + self.test_cfg.center_size
            z_feat_center.append(z_feat[i][:, :, left:right, left:right])
        return tuple(z_feat_center)

    def forward_search(self, x_img: Tensor) -> Tuple[Tensor, ...]:
        """Extract the features of search images.

        Args:
            x_img (Tensor): of shape (N, C, H, W) encoding input search
                images. Typically H and W equal to 255.

        Returns:
            Tuple[Tensor, ...]: Multi level feature map of search images.
        """
        x_feat = self.backbone(x_img)
        if self.with_neck:
            x_feat = self.neck(x_feat)
        return x_feat

    def get_cropped_img(self, img: Tensor, center_xy: Tensor,
                        target_size: Tensor, crop_size: Tensor,
                        avg_channel: Tensor) -> Tensor:
        """Crop image.

        Only used during testing.

        This function mainly contains two steps:
        1. Crop `img` based on center `center_xy` and size `crop_size`. If the
        cropped image is out of boundary of `img`, use `avg_channel` to pad.
        2. Resize the cropped image to `target_size`.

        Args:
            img (Tensor): of shape (1, C, H, W) encoding original input
                image.
            center_xy (Tensor): of shape (2, ) denoting the center point
                for cropping image.
            target_size (int): The output size of cropped image.
            crop_size (Tensor): The size for cropping image.
            avg_channel (Tensor): of shape (3, ) denoting the padding
                values.

        Returns:
            Tensor: of shape (1, C, target_size, target_size) encoding
                the resized cropped image.
        """
        N, C, H, W = img.shape
        context_xmin = int(center_xy[0] - crop_size / 2)
        context_xmax = int(center_xy[0] + crop_size / 2)
        context_ymin = int(center_xy[1] - crop_size / 2)
        context_ymax = int(center_xy[1] + crop_size / 2)

        left_pad = max(0, -context_xmin)
        top_pad = max(0, -context_ymin)
        right_pad = max(0, context_xmax - W)
        bottom_pad = max(0, context_ymax - H)

        context_xmin += left_pad
        context_xmax += left_pad
        context_ymin += top_pad
        context_ymax += top_pad

        avg_channel = avg_channel[:, None, None]
        if any([top_pad, bottom_pad, left_pad, right_pad]):
            new_img = img.new_zeros(N, C, H + top_pad + bottom_pad,
                                    W + left_pad + right_pad)
            new_img[..., top_pad:top_pad + H, left_pad:left_pad + W] = img
            if top_pad:
                new_img[..., :top_pad, left_pad:left_pad + W] = avg_channel
            if bottom_pad:
                new_img[..., H + top_pad:, left_pad:left_pad + W] = avg_channel
            if left_pad:
                new_img[..., :left_pad] = avg_channel
            if right_pad:
                new_img[..., W + left_pad:] = avg_channel
            crop_img = new_img[..., context_ymin:context_ymax + 1,
                               context_xmin:context_xmax + 1]
        else:
            crop_img = img[..., context_ymin:context_ymax + 1,
                           context_xmin:context_xmax + 1]

        crop_img = torch.nn.functional.interpolate(
            crop_img,
            size=(target_size, target_size),
            mode='bilinear',
            align_corners=False)
        return crop_img

    def init(self, img: Tensor) -> Tuple[Tuple[Tensor, ...], Tensor]:
        """Initialize the single object tracker in the first frame.

        Args:
            img (Tensor): of shape (1, C, H, W) encoding original input
                image.
        """
        bbox = self.memo.bbox
        z_width = bbox[2] + self.test_cfg.context_amount * (bbox[2] + bbox[3])
        z_height = bbox[3] + self.test_cfg.context_amount * (bbox[2] + bbox[3])
        z_size = torch.round(torch.sqrt(z_width * z_height))
        # used for padding when cropping the image.
        avg_channel = torch.mean(img, dim=(0, 2, 3))
        z_crop = self.get_cropped_img(img, bbox[0:2],
                                      self.test_cfg.exemplar_size, z_size,
                                      avg_channel)
        self.memo.z_feat = self.forward_template(z_crop)
        self.memo.avg_channel = avg_channel

    def track(self, img: Tensor, data_samples: SampleList) -> InstanceList:
        """Track the box of previous frame to current frame `img`.

        Args:
            img (Tensor): of shape (1, C, H, W) encoding original input
                image.
            data_samples (list[:obj:`TrackDataSample`]): The batch
                data samples. It usually includes information such
                as ``gt_instances`` and 'metainfo'.

        Returns:
            InstanceList: Tracking results of each image after the postprocess.
                - scores: a Tensor denoting the score of best_bbox.
                - bboxes: a Tensor of shape (4, ) in [x1, x2, y1, y2]
                format, and denotes the best tracked bbox in current frame.
        """
        prev_bbox = self.memo.bbox
        z_width = prev_bbox[2] + self.test_cfg.context_amount * (
            prev_bbox[2] + prev_bbox[3])
        z_height = prev_bbox[3] + self.test_cfg.context_amount * (
            prev_bbox[2] + prev_bbox[3])
        z_size = torch.sqrt(z_width * z_height)

        x_size = torch.round(
            z_size * (self.test_cfg.search_size / self.test_cfg.exemplar_size))
        x_crop = self.get_cropped_img(img, prev_bbox[0:2],
                                      self.test_cfg.search_size, x_size,
                                      self.memo.avg_channel)

        x_feat = self.forward_search(x_crop)
        scale_factor = self.test_cfg.exemplar_size / z_size

        results = self.head.predict(self.memo.z_feat, x_feat, data_samples,
                                    prev_bbox, scale_factor)

        return results

    def loss(self, inputs: dict, data_samples: SampleList, **kwargs) -> dict:
        """
        Args:
            inputs (Dict[str, Tensor]): of shape (N, T, C, H, W) encoding
                input images. Typically these should be mean centered and std
                scaled. The N denotes batch size. The T denotes the number of
                key/reference frames.
                - img (Tensor) : The key images.
                - ref_img (Tensor): The reference images.

            data_samples (list[:obj:`TrackDataSample`]): The batch
                data samples. It usually includes information such
                as ``gt_instance``.

        Return:
            dict: A dictionary of loss components.
        """
        search_img = inputs['search_img']
        assert search_img.dim(
        ) == 5, 'The img must be 5D Tensor (N, T, C, H, W).'
        search_img = search_img[:, 0]

        template_img = inputs['img']
        assert template_img.dim(
        ) == 5, 'The img must be 5D Tensor (N, T, C, H, W).'
        template_img = template_img[:, 0]

        z_feat = self.forward_template(template_img)
        x_feat = self.forward_search(search_img)
        losses = self.head.loss(z_feat, x_feat, data_samples, **kwargs)
        return losses



# model settings
_base_ = [
    './qdtrack_faster-rcnn_r50_fpn_4e_base.py', '../../_base_/datasets/tao.py'
]
model = dict(
    type='QDTrack',
    data_preprocessor=dict(
        mean=[123.675, 116.28, 103.53],
        std=[58.395, 57.12, 57.375],
        bgr_to_rgb=True),
    detector=dict(
        backbone=dict(
            depth=101,
            norm_cfg=dict(requires_grad=True),
            style='pytorch',
            init_cfg=dict(
                type='Pretrained', checkpoint='torchvision://resnet101')),
        rpn_head=dict(bbox_coder=dict(clip_border=True)),
        roi_head=dict(
            bbox_head=dict(bbox_coder=dict(
                clip_border=True), num_classes=482)),
        test_cfg=dict(
            rcnn=dict(
                score_thr=0.0001,
                nms=dict(type='nms', iou_threshold=0.5),
                max_per_img=300)),
        init_cfg=None),
    track_head=dict(train_cfg=dict(assigner=dict(neg_iou_thr=0.3))),
    tracker=dict(
        _delete_=True,
        type='QuasiDenseTAOTracker',
        init_score_thr=0.0001,
        obj_score_thr=0.0001,
        match_score_thr=0.5,
        memo_frames=10,
        memo_momentum=0.8,
        momentum_obj_score=0.5,
        obj_score_diff_thr=1.0,
        distractor_nms_thr=0.3,
        distractor_score_thr=0.5,
        match_metric='bisoftmax',
        match_with_cosine=True))

# learning policy
param_scheduler = [
    dict(
        type='LinearLR',
        start_factor=1.0 / 1000,
        by_epoch=False,
        begin=0,
        end=1000),
    dict(
        type='mmdet.MultiStepLR',
        begin=0,
        end=24,
        by_epoch=True,
        milestones=[16, 22])
]

# optimizer
optim_wrapper = dict(
    type='OptimWrapper',
    optimizer=dict(type='SGD', lr=0.02, momentum=0.9, weight_decay=0.0001),
    clip_grad=None)

# runtime settings
train_cfg = dict(
    type='EpochBasedTrainLoop', max_epochs=24, val_begin=16, val_interval=2)
val_cfg = dict(type='ValLoop')
test_cfg = dict(type='TestLoop')

# evaluator
val_evaluator = dict(type='TAOMetric', metric=['bbox'])
test_evaluator = val_evaluator



# Copyright (c) OpenMMLab. All rights reserved.
from typing import Iterator, Sized

import numpy as np
from mmengine.dist import get_dist_info
from torch.utils.data import Sampler

from mmtrack.datasets import BaseSOTDataset, BaseVideoDataset
from mmtrack.registry import DATA_SAMPLERS


@DATA_SAMPLERS.register_module()
class VideoSampler(Sampler):
    """The video data sampler is for both distributed and non-distributed
    environment. It is only used in testing.

    Args:
        dataset (Sized): The dataset.
    """

    def __init__(self, dataset: Sized, seed: int = 0) -> None:
        self.dataset = dataset
        assert self.dataset.test_mode

        rank, world_size = get_dist_info()
        self.rank = rank
        self.world_size = world_size

        if isinstance(self.dataset, BaseSOTDataset):
            # The input of '__getitem__' function in SOT dataset class must be
            # a tuple when testing. The tuple is in (video_index, frame_index)
            # format.
            self.num_videos = self.dataset.num_videos
            if self.num_videos < self.world_size:
                raise ValueError(f'only {self.num_videos} videos loaded,'
                                 f'but {self.world_size} gpus were given.')

            chunks = np.array_split(
                list(range(self.num_videos)), self.world_size)
            self.indices = []
            for videos in chunks:
                indices_chunk = []
                for video_ind in videos:
                    indices_chunk.extend([
                        (video_ind, frame_ind) for frame_ind in range(
                            self.dataset.get_len_per_video(video_ind))
                    ])
                self.indices.append(indices_chunk)
        else:
            assert isinstance(self.dataset, BaseVideoDataset)
            first_frame_indices = []
            for i in range(len(self.dataset)):
                data_info = self.dataset.get_data_info(i)
                if data_info['frame_id'] == 0:
                    first_frame_indices.append(i)

            self.num_videos = len(first_frame_indices)
            if self.num_videos < self.world_size:
                raise ValueError(f'only {self.num_videos} videos loaded,'
                                 f'but {self.world_size} gpus were given.')

            chunks = np.array_split(first_frame_indices, self.world_size)
            split_flags = [c[0] for c in chunks]
            split_flags.append(len(self.dataset))

            self.indices = [
                list(range(split_flags[i], split_flags[i + 1]))
                for i in range(self.world_size)
            ]

    def __iter__(self) -> Iterator[int]:
        """Iterate the indices."""
        indices = self.indices[self.rank]
        return iter(indices)

    def __len__(self) -> int:
        """The number of samples in this rank."""
        return len(self.indices[self.rank])

    def set_epoch(self, epoch: int) -> None:
        """Not supported in iteration-based runner."""
        raise NotImplementedError(
            'The `VideoSampler` is only used in testing, '
            "and doesn't need `set_epoch`")



# Copyright (c) OpenMMLab. All rights reserved.
from typing import List, Optional, Tuple

import torch
import torch.nn as nn
from mmdet.models.task_modules import SamplingResult
from torch import Tensor

from mmtrack.registry import MODELS
from ..task_modules.track import embed_similarity
from .roi_embed_head import RoIEmbedHead


@MODELS.register_module()
class QuasiDenseEmbedHead(RoIEmbedHead):
    """The quasi-dense roi embed head.

    Args:
        embed_channels (int): The input channel of embed features.
            Defaults to 256.
        softmax_temp (int): Softmax temperature. Defaults to -1.
        loss_track (dict): The loss function for tracking. Defaults to
            MultiPosCrossEntropyLoss.
        loss_track_aux (dict): The auxiliary loss function for tracking.
            Defaults to L2Loss.
    """

    def __init__(self,
                 embed_channels: int = 256,
                 softmax_temp: int = -1,
                 loss_track: Optional[dict] = None,
                 loss_track_aux: dict = dict(
                     type='L2Loss',
                     sample_ratio=3,
                     margin=0.3,
                     loss_weight=1.0,
                     hard_mining=True),
                 init_cfg: dict = dict(
                     type='Xavier',
                     layer='Linear',
                     distribution='uniform',
                     bias=0,
                     override=dict(
                         type='Normal',
                         name='fc_embed',
                         mean=0,
                         std=0.01,
                         bias=0)),
                 **kwargs):
        super(QuasiDenseEmbedHead, self).__init__(init_cfg=init_cfg, **kwargs)

        if loss_track is None:
            loss_track = dict(
                type='MultiPosCrossEntropyLoss', loss_weight=0.25)

        self.fc_embed = nn.Linear(self.last_layer_dim, embed_channels)
        self.softmax_temp = softmax_temp
        self.loss_track = MODELS.build(loss_track)
        if loss_track_aux is not None:
            self.loss_track_aux = MODELS.build(loss_track_aux)
        else:
            self.loss_track_aux = None

    def forward(self, x: Tensor) -> Tensor:
        """Forward the input `x`."""

        if self.num_convs > 0:
            for conv in self.convs:
                x = conv(x)
        x = x.flatten(1)
        if self.num_fcs > 0:
            for fc in self.fcs:
                x = self.relu(fc(x))
        x = self.fc_embed(x)
        return x

    def get_targets(
            self, gt_match_indices: List[Tensor],
            key_sampling_results: List[SamplingResult],
            ref_sampling_results: List[SamplingResult]) -> Tuple[List, List]:
        """Calculate the track targets and track weights for all samples in a
        batch according to the sampling_results.

        Args:
            gt_match_indices (list(Tensor)): Mapping from gt_instance_ids to
                ref_gt_instance_ids of the same tracklet in a pair of images.
            key_sampling_results (List[obj:SamplingResult]): Assign results of
                all images in a batch after sampling.
            ref_sampling_results (List[obj:SamplingResult]): Assign results of
                all reference images in a batch after sampling.

        Returns:
            Tuple[list[Tensor]]: Association results.
            Containing the following list of Tensors:

                - track_targets (list[Tensor]): The mapping instance ids from
                    all positive proposals in the key image to all proposals
                    in the reference image, each tensor in list has
                    shape (len(key_pos_bboxes), len(ref_bboxes)).
                - track_weights (list[Tensor]): Loss weights for all positive
                    proposals in a batch, each tensor in list has
                    shape (len(key_pos_bboxes),).
        """

        track_targets = []
        track_weights = []
        for _gt_match_indices, key_res, ref_res in zip(gt_match_indices,
                                                       key_sampling_results,
                                                       ref_sampling_results):
            targets = _gt_match_indices.new_zeros(
                (key_res.pos_bboxes.size(0), ref_res.bboxes.size(0)),
                dtype=torch.int)
            _match_indices = _gt_match_indices[key_res.pos_assigned_gt_inds]
            pos2pos = (_match_indices.view(
                -1, 1) == ref_res.pos_assigned_gt_inds.view(1, -1)).int()
            targets[:, :pos2pos.size(1)] = pos2pos
            weights = (targets.sum(dim=1) > 0).float()
            track_targets.append(targets)
            track_weights.append(weights)
        return track_targets, track_weights

    def match(
        self, key_embeds: Tensor, ref_embeds: Tensor,
        key_sampling_results: List[SamplingResult],
        ref_sampling_results: List[SamplingResult]
    ) -> Tuple[List[Tensor], List[Tensor]]:
        """Calculate the dist matrixes for loss measurement.

        Args:
            key_embeds (Tensor): Embeds of positive bboxes in sampling results
                of key image.
            ref_embeds (Tensor): Embeds of all bboxes in sampling results
                of the reference image.
            key_sampling_results (List[obj:SamplingResults]): Assign results of
                all images in a batch after sampling.
            ref_sampling_results (List[obj:SamplingResults]): Assign results of
                all reference images in a batch after sampling.

        Returns:
            Tuple[list[Tensor]]: Calculation results.
            Containing the following list of Tensors:

                - dists (list[Tensor]): Dot-product dists between
                    key_embeds and ref_embeds, each tensor in list has
                    shape (len(key_pos_bboxes), len(ref_bboxes)).
                - cos_dists (list[Tensor]): Cosine dists between
                    key_embeds and ref_embeds, each tensor in list has
                    shape (len(key_pos_bboxes), len(ref_bboxes)).
        """

        num_key_rois = [res.pos_bboxes.size(0) for res in key_sampling_results]
        key_embeds = torch.split(key_embeds, num_key_rois)
        num_ref_rois = [res.bboxes.size(0) for res in ref_sampling_results]
        ref_embeds = torch.split(ref_embeds, num_ref_rois)

        dists, cos_dists = [], []
        for key_embed, ref_embed in zip(key_embeds, ref_embeds):
            dist = embed_similarity(
                key_embed,
                ref_embed,
                method='dot_product',
                temperature=self.softmax_temp)
            dists.append(dist)
            if self.loss_track_aux is not None:
                cos_dist = embed_similarity(
                    key_embed, ref_embed, method='cosine')
                cos_dists.append(cos_dist)
            else:
                cos_dists.append(None)
        return dists, cos_dists

    def loss(self, key_roi_feats: Tensor, ref_roi_feats: Tensor,
             key_sampling_results: List[SamplingResult],
             ref_sampling_results: List[SamplingResult],
             gt_match_indices_list: List[Tensor]) -> dict:
        """Calculate the track loss and the auxiliary track loss.

        Args:
            key_roi_feats (Tensor): Embeds of positive bboxes in sampling
                results of key image.
            ref_roi_feats (Tensor): Embeds of all bboxes in sampling results
                of the reference image.
            key_sampling_results (List[obj:SamplingResults]): Assign results of
                all images in a batch after sampling.
            ref_sampling_results (List[obj:SamplingResults]): Assign results of
                all reference images in a batch after sampling.
            gt_match_indices_list (list(Tensor)): Mapping from gt_instances_id
                to ref_gt_instances_id of the same tracklet in a pair of
                images.

        Returns:
            Dict [str: Tensor]: Calculation results.
            Containing the following list of Tensors:

                - loss_track (Tensor): Results of loss_track function.
                - loss_track_aux (Tensor): Results of loss_track_aux function.
        """
        key_track_feats = self(key_roi_feats)
        ref_track_feats = self(ref_roi_feats)

        losses = self.loss_by_feat(key_track_feats, ref_track_feats,
                                   key_sampling_results, ref_sampling_results,
                                   gt_match_indices_list)
        return losses

    def loss_by_feat(self, key_track_feats: Tensor, ref_track_feats: Tensor,
                     key_sampling_results: List[SamplingResult],
                     ref_sampling_results: List[SamplingResult],
                     gt_match_indices_list: List[Tensor]) -> dict:
        """Calculate the track loss and the auxiliary track loss.

        Args:
            key_track_feats (Tensor): Embeds of positive bboxes in sampling
                results of key image.
            ref_track_feats (Tensor): Embeds of all bboxes in sampling results
                of the reference image.
            key_sampling_results (List[obj:SamplingResults]): Assign results of
                all images in a batch after sampling.
            ref_sampling_results (List[obj:SamplingResults]): Assign results of
                all reference images in a batch after sampling.
            gt_match_indices_list (list(Tensor)): Mapping from gt_instances_id
                to ref_gt_instances_id of the same tracklet in a pair of
                images.

        Returns:
            Dict [str: Tensor]: Calculation results.
            Containing the following list of Tensors:

                - loss_track (Tensor): Results of loss_track function.
                - loss_track_aux (Tensor): Results of loss_track_aux function.
        """
        dists, cos_dists = self.match(key_track_feats, ref_track_feats,
                                      key_sampling_results,
                                      ref_sampling_results)
        targets, weights = self.get_targets(gt_match_indices_list,
                                            key_sampling_results,
                                            ref_sampling_results)
        losses = dict()

        loss_track = 0.
        loss_track_aux = 0.
        for _dists, _cos_dists, _targets, _weights in zip(
                dists, cos_dists, targets, weights):
            loss_track += self.loss_track(
                _dists, _targets, _weights, avg_factor=_weights.sum())
            if self.loss_track_aux is not None:
                loss_track_aux += self.loss_track_aux(_cos_dists, _targets)
        losses['loss_track'] = loss_track / len(dists)

        if self.loss_track_aux is not None:
            losses['loss_track_aux'] = loss_track_aux / len(dists)

        return losses

    def predict(self, bbox_feats: Tensor) -> Tensor:
        """Perform forward propagation of the tracking head and predict
        tracking results on the features of the upstream network.

        Args:
            bbox_feats: The extracted roi features.

        Returns:
            Tensor: The extracted track features.
        """
        track_feats = self(bbox_feats)
        return track_feats



# Copyright (c) OpenMMLab. All rights reserved.
from .inference import inference_mot, inference_sot, inference_vid, init_model

__all__ = ['init_model', 'inference_mot', 'inference_sot', 'inference_vid']



# Copyright (c) OpenMMLab. All rights reserved.
import argparse
import glob
import os
import os.path as osp
import re
from collections import defaultdict

import mmcv
import mmengine
from tqdm import tqdm


def parse_args():
    parser = argparse.ArgumentParser(
        description='OTB100 dataset to COCO Video format')
    parser.add_argument(
        '-i',
        '--input',
        help='root directory of OTB100 dataset',
    )
    parser.add_argument(
        '-o',
        '--output',
        help='directory to save coco formatted label file',
    )
    return parser.parse_args()


def convert_otb100(otb, ann_dir, save_dir):
    """Convert OTB100 dataset to COCO style.

    Args:
        otb (dict): The converted COCO style annotations.
        ann_dir (str): The path of OTB100 dataset
        save_dir (str): The path to save `OTB100`.
    """
    records = dict(vid_id=1, img_id=1, ann_id=1, global_instance_id=1)
    ann_dir = osp.join(ann_dir, 'data')
    videos_list = os.listdir(ann_dir)
    otb['categories'] = [dict(id=0, name=0)]

    for video_name in tqdm(videos_list):
        video_path = osp.join(ann_dir, video_name)

        if video_name == 'David':
            start_frame_id = 300
        # The first five frames in Tiger1 can not be used
        # as the initinal frames.
        # Details can be seen in tracker_benchmark_v1.0/initOmit/tiger1.txt.
        # The start_frame_id is 1-based.
        elif video_name == 'Tiger1':
            start_frame_id = 6
        else:
            start_frame_id = 1

        img_list = os.listdir(osp.join(video_path, 'img'))
        img_list = sorted(img_list)

        img = mmcv.imread(osp.join(video_path, 'img', img_list[0]))
        height, width, _ = img.shape

        # One video may have several tracking instances with their
        # respective annotations.
        gt_list = glob.glob(
            osp.join(ann_dir, video_name, 'groundtruth_rect*.txt'))
        for gt_file in gt_list:
            # exclude empty files
            if osp.getsize(gt_file) == 0:
                continue

            video = dict(id=records['vid_id'], name=video_name)
            otb['videos'].append(video)

            gt_bboxes = mmengine.list_from_file(gt_file)
            if video_name == 'Tiger1':
                gt_bboxes = gt_bboxes[start_frame_id - 1:]
            for frame_id, gt_bbox in enumerate(gt_bboxes):
                src_frame_id = frame_id + start_frame_id - 1
                file_name = osp.join(video_name, 'img', img_list[src_frame_id])
                image = dict(
                    file_name=file_name,
                    height=height,
                    width=width,
                    id=records['img_id'],
                    frame_id=frame_id,
                    video_id=records['vid_id'])
                otb['images'].append(image)

                bbox = list(map(int, re.findall(r'-?\d+', gt_bbox)))
                assert len(bbox) == 4
                anno_dict = dict(
                    id=records['ann_id'],
                    video_id=records['vid_id'],
                    image_id=records['img_id'],
                    instance_id=records['global_instance_id'],
                    category_id=0,
                    bbox=bbox,
                    area=bbox[2] * bbox[3],
                )
                otb['annotations'].append(anno_dict)

                records['ann_id'] += 1
                records['img_id'] += 1

            records['global_instance_id'] += 1
            records['vid_id'] += 1

    if not osp.isdir(save_dir):
        os.makedirs(save_dir)
    mmengine.dump(otb, osp.join(save_dir, 'otb100.json'))
    print('-----OTB100 Dataset------')
    print(f'{records["vid_id"]- 1} videos')
    print(f'{records["global_instance_id"]- 1} instances')
    print(f'{records["img_id"]- 1} images')
    print(f'{records["ann_id"] - 1} objects')
    print('-----------------------------')


def main():
    args = parse_args()
    otb = defaultdict(list)
    convert_otb100(otb, args.input, args.output)


if __name__ == '__main__':
    main()



# Copyright (c) OpenMMLab. All rights reserved.
import copy
import logging
import os
import tempfile
import warnings
from typing import Optional, Union

import mmengine
import numpy as np
import torch
from mmcv.ops import RoIPool
from mmengine.dataset import Compose
from mmengine.logging import MMLogger
from mmengine.runner import load_checkpoint
from torch import nn

from mmtrack.registry import MODELS
from mmtrack.utils import SampleList


def init_model(config: Union[str, mmengine.Config],
               checkpoint: Optional[str] = None,
               device: str = 'cuda:0',
               cfg_options: Optional[dict] = None,
               verbose_init_params: bool = False) -> nn.Module:
    """Initialize a model from config file.

    Args:
        config (str or :obj:`mmengine.Config`): Config file path or the config
            object.
        checkpoint (Optional[str], optional): Checkpoint path. Defaults to
            None.
        device (str, optional): The device that the model inferences on.
            Defaults to `cuda:0`.
        cfg_options (Optional[dict], optional): Options to override some
            settings in the used config. Defaults to None.
        verbose_init_params (bool, optional): Whether to print the information
            of initialized parameters to the console. Defaults to False.

    Returns:
        nn.Module: The constructed model.
    """
    if isinstance(config, str):
        config = mmengine.Config.fromfile(config)
    elif not isinstance(config, mmengine.Config):
        raise TypeError('config must be a filename or Config object, '
                        f'but got {type(config)}')
    if cfg_options is not None:
        config.merge_from_dict(cfg_options)
    model = MODELS.build(config.model)

    if not verbose_init_params:
        # Creating a temporary file to record the information of initialized
        # parameters. If not, the information of initialized parameters will be
        # printed to the console because of the call of
        # `mmcv.runner.BaseModule.init_weights`.
        tmp_file = tempfile.NamedTemporaryFile(delete=False)
        file_handler = logging.FileHandler(tmp_file.name, mode='w')
        logger = MMLogger.get_current_instance()
        logger.addHandler(file_handler)
        # We need call `init_weights()` to load pretained weights in MOT
        # task.
        model.init_weights()
        file_handler.close()
        logger.removeHandler(file_handler)
        tmp_file.close()
        os.remove(tmp_file.name)
    else:
        # We need call `init_weights()` to load pretained weights in MOT task.
        model.init_weights()

    if checkpoint is not None:
        checkpoint = load_checkpoint(model, checkpoint, map_location='cpu')
        # Weights converted from elsewhere may not have meta fields.
        checkpoint_meta = checkpoint.get('meta', {})
        # save the dataset_meta in the model for convenience
        if 'dataset_meta' in checkpoint_meta:
            # mmtrack 1.x
            model.dataset_meta = checkpoint_meta['dataset_meta']
        elif 'CLASSES' in checkpoint_meta:
            # < mmtrack 1.x
            classes = checkpoint_meta['CLASSES']
            model.dataset_meta = {'CLASSES': classes}

    # Some methods don't load checkpoints or checkpoints don't contain
    # 'dataset_meta'
    if not hasattr(model, 'dataset_meta'):
        warnings.simplefilter('once')
        warnings.warn('dataset_meta or class names are missed, '
                      'use None by default.')
        model.dataset_meta = {'CLASSES': None}

    model.cfg = config  # save the config in the model for convenience
    model.to(device)
    model.eval()
    return model


def inference_mot(model: nn.Module, img: np.ndarray,
                  frame_id: int) -> SampleList:
    """Inference image(s) with the mot model.

    Args:
        model (nn.Module): The loaded mot model.
        img (np.ndarray): Loaded image.
        frame_id (int): frame id.

    Returns:
        SampleList: The tracking data samples.
    """
    cfg = model.cfg
    data = dict(
        img=img.astype(np.float32), frame_id=frame_id, ori_shape=img.shape[:2])
    # remove the "LoadImageFromFile" and "LoadTrackAnnotations" in pipeline
    test_pipeline = Compose(cfg.test_dataloader.dataset.pipeline[2:])
    data = test_pipeline(data)

    if not next(model.parameters()).is_cuda:
        for m in model.modules():
            assert not isinstance(
                m, RoIPool
            ), 'CPU inference with RoIPool is not supported currently.'

    # forward the model
    with torch.no_grad():
        data = mmengine.dataset.default_collate([data])
        result = model.test_step(data)[0]
    return result


def inference_sot(model: nn.Module, image: np.ndarray, init_bbox: np.ndarray,
                  frame_id: int) -> SampleList:
    """Inference image with the single object tracker.

    Args:
        model (nn.Module): The loaded tracker.
        image (np.ndarray): Loaded images.
        init_bbox (np.ndarray): The target needs to be tracked.
        frame_id (int): frame id.

    Returns:
        SampleList: The tracking data samples.
    """
    cfg = model.cfg
    data = dict(
        img=image.astype(np.float32),
        gt_bboxes=np.array(init_bbox).astype(np.float32),
        frame_id=frame_id,
        ori_shape=image.shape[:2])
    # remove the "LoadImageFromFile" and "LoadAnnotations" in pipeline
    test_pipeline = Compose(cfg.test_dataloader.dataset.pipeline[2:])
    data = test_pipeline(data)

    if not next(model.parameters()).is_cuda:
        for m in model.modules():
            assert not isinstance(
                m, RoIPool
            ), 'CPU inference with RoIPool is not supported currently.'

    # forward the model
    with torch.no_grad():
        data = mmengine.dataset.default_collate([data])
        result = model.test_step(data)[0]
    return result


def inference_vid(
    model: nn.Module,
    image: np.ndarray,
    frame_id: int,
    ref_img_sampler: dict = dict(frame_stride=2, num_left_ref_imgs=10)
) -> SampleList:
    """Inference image with the video object detector.

    Args:
        model (nn.Module): The loaded detector.
        image (ndarray): Loaded images.
        frame_id (int): Frame id.
        ref_img_sampler (dict): The configuration for sampling reference
            images. Only used under video detector of fgfa style. Defaults to
            dict(frame_stride=2, num_left_ref_imgs=10).

    Returns:
        SampleList: The detection results.
    """
    cfg = model.cfg

    first_transform = cfg.test_dataloader.dataset.pipeline[0]
    if first_transform.type == 'LoadImageFromFile':
        data = dict(
            img=image.astype(np.float32).copy(),
            frame_id=frame_id,
            ori_shape=image.shape[:2])
        # remove the "LoadImageFromFile" in pipeline
        test_pipeline = Compose(cfg.test_dataloader.dataset.pipeline[1:])
    elif first_transform.type == 'TransformBroadcaster':
        assert first_transform.transforms[0].type == 'LoadImageFromFile'
        # Only used under video detector of fgfa style.
        data = dict(
            img=[image.astype(np.float32).copy()],
            frame_id=[frame_id],
            ori_shape=[image.shape[:2]])

        num_left_ref_imgs = ref_img_sampler.get('num_left_ref_imgs')
        frame_stride = ref_img_sampler.get('frame_stride')
        if frame_id == 0:
            for i in range(num_left_ref_imgs):
                data['img'].append(image.astype(np.float32).copy())
                data['frame_id'].append(frame_id)
                data['ori_shape'].append(image.shape[:2])
        elif frame_id % frame_stride == 0:
            data['img'].append(image.astype(np.float32).copy())
            data['frame_id'].append(frame_id)
            data['ori_shape'].append(image.shape[:2])
        # In order to pop the LoadImageFromFile, test_pipeline[0] is
        # `TransformBroadcaster` and test_pipeline[0].transforms[0]
        # is 'LoadImageFromFile'.
        test_pipeline = copy.deepcopy(cfg.test_dataloader.dataset.pipeline)
        test_pipeline[0].transforms.pop(0)
        test_pipeline = Compose(test_pipeline)
    else:
        print('Not supported loading data pipeline type: '
              f'{first_transform.type}')
        raise NotImplementedError

    data = test_pipeline(data)

    if not next(model.parameters()).is_cuda:
        for m in model.modules():
            assert not isinstance(
                m, RoIPool
            ), 'CPU inference with RoIPool is not supported currently.'

    # forward the model
    with torch.no_grad():
        data = mmengine.dataset.default_collate([data])
        result = model.test_step(data)[0]
    return result



# Copyright (c) OpenMMLab. All rights reserved.
import os.path as osp
from typing import List

import numpy as np
from mmengine.dataset import force_full_init
from mmengine.fileio.file_client import FileClient

from mmtrack.registry import DATASETS
from .api_wrappers import CocoVID
from .base_sot_dataset import BaseSOTDataset


@DATASETS.register_module()
class SOTImageNetVIDDataset(BaseSOTDataset):
    """ImageNet VID dataset of single object tracking.

    The dataset only support training mode.
    """

    def __init__(self, *args, **kwargs):
        """Initialization of SOT dataset class."""
        super().__init__(*args, **kwargs)

    def load_data_list(self) -> List[dict]:
        """Load annotations from an annotation file named as ``self.ann_file``.

        Returns:
            list[dict]: The length of the list is the number of instances. The
                inner dict contains instance ID in CocoVID API.
        """
        file_client = FileClient.infer_client(uri=self.ann_file)
        with file_client.get_local_path(self.ann_file) as local_file:
            self.coco = CocoVID(local_file)
        data_infos = [
            dict(ins_id=ins_id) for ins_id in self.coco.instancesToImgs.keys()
        ]
        return data_infos

    def get_bboxes_from_video(self, video_idx: int) -> np.ndarray:
        """Get bbox annotation about one instance in a video. Considering
        `get_bboxes_from_video` in `SOTBaseDataset` is not compatible with
        `SOTImageNetVIDDataset`, we oveload this function though it's not
        called by `self.get_ann_infos_from_video`.

        Args:
            video_idx (int): The index of video. Here, each video_idx denotes
                an instance.

        Returns:
            ndarray: In [N, 4] shape. The bbox is in (x, y, w, h) format.
        """
        instance_id = self.get_data_info(video_idx)['ins_id']
        img_ids = self.coco.instancesToImgs[instance_id]
        bboxes = []
        for img_id in img_ids:
            for ann in self.coco.imgToAnns[img_id]:
                if ann['instance_id'] == instance_id:
                    bboxes.append(ann['bbox'])
        bboxes = np.array(bboxes, dtype=np.float32).reshape(-1, 4)
        return bboxes

    def get_img_infos_from_video(self, video_idx: int) -> dict:
        """Get image information about one instance in a video.

        Args:
            video_idx (int): The index of video.

        Returns:
            dict: {
                    'video_id': int,
                    'frame_ids': np.ndarray,
                    'img_paths': list[str],
                    'video_length': int
                  }
        """
        instance_id = self.get_data_info(video_idx)['ins_id']
        img_ids = self.coco.instancesToImgs[instance_id]
        frame_ids = []
        img_names = []
        # In ImageNetVID dataset, frame_ids may not be continuous.
        for img_id in img_ids:
            frame_ids.append(self.coco.imgs[img_id]['frame_id'])
            img_names.append(
                osp.join(self.data_prefix['img_path'],
                         self.coco.imgs[img_id]['file_name']))
        img_infos = dict(
            video_id=video_idx,
            frame_ids=frame_ids,
            img_paths=img_names,
            video_length=len(frame_ids))
        return img_infos

    def get_ann_infos_from_video(self, video_idx: int) -> dict:
        """Get annotation information about one instance in a video.
        Note: We overload this function for speed up loading video information.

        Args:
            video_idx (int): The index of video. Here, each video_idx denotes
                an instance.

        Returns:
            dict: {
                    'bboxes': np.ndarray in (N, 4) shape,
                    'bboxes_isvalid': np.ndarray,
                    'visible': np.ndarray
                  }.
                  The annotation information in some datasets may contain
                    'visible_ratio'. The bbox is in (x1, y1, x2, y2) format.
        """
        instance_id = self.get_data_info(video_idx)['ins_id']
        img_ids = self.coco.instancesToImgs[instance_id]
        bboxes = []
        visible = []
        for img_id in img_ids:
            for ann in self.coco.imgToAnns[img_id]:
                if ann['instance_id'] == instance_id:
                    bboxes.append(ann['bbox'])
                    visible.append(not ann.get('occluded', False))
        bboxes = np.array(bboxes, dtype=np.float32).reshape(-1, 4)
        bboxes_isvalid = (bboxes[:, 2] > self.bbox_min_size) & (
            bboxes[:, 3] > self.bbox_min_size)
        bboxes[:, 2:] += bboxes[:, :2]
        visible = np.array(visible, dtype=np.bool_) & bboxes_isvalid
        ann_infos = ann_infos = dict(
            bboxes=bboxes, bboxes_isvalid=bboxes_isvalid, visible=visible)
        return ann_infos

    def get_visibility_from_video(self, video_idx: int) -> dict:
        """Get the visible information about one instance in a video.
        Considering `get_visibility_from_video` in `SOTBaseDataset` is not
        compatible with `SOTImageNetVIDDataset`, we oveload this function
        though it's not called by `self.get_ann_infos_from_video`.

        Args:
            video_idx (int): The index of video.

        Returns:
            dict: The visibilities of each object in the video.
        """
        instance_id = self.get_data_info(video_idx)['ins_id']
        img_ids = self.coco.instancesToImgs[instance_id]
        visible = []
        for img_id in img_ids:
            for ann in self.coco.imgToAnns[img_id]:
                if ann['instance_id'] == instance_id:
                    visible.append(not ann.get('occluded', False))
        visible_info = dict(visible=np.array(visible, dtype=np.bool_))
        return visible_info

    @force_full_init
    def get_len_per_video(self, video_idx: int) -> int:
        """Get the length of filtered dataset and automatically call
        ``full_init`` if the  dataset has not been fully init.

        Returns:
            int: The length of filtered dataset.
        """
        instance_id = self.get_data_info(video_idx)['ins_id']
        return len(self.coco.instancesToImgs[instance_id])



_base_ = [
    '../../_base_/datasets/trackingnet.py',
    './prdimp_r50_8xb10-50e_got10k-lasot-trackingnet-coco_base.py'
]

# evaluator
val_evaluator = dict(outfile_prefix='results/prdimp_trackingnet')
test_evaluator = val_evaluator



# test pipeline
test_pipeline = [
    dict(type='LoadImageFromFile'),
    dict(type='LoadTrackAnnotations', with_instance_id=False),
    dict(type='PackTrackInputs', pack_single_img=True)
]

# dataloader
val_dataloader = dict(
    batch_size=1,
    num_workers=4,
    persistent_workers=True,
    drop_last=False,
    sampler=dict(type='VideoSampler'),
    dataset=dict(
        type='GOT10kDataset',
        data_root='data/',
        ann_file='GOT10k/annotations/got10k_test_infos.txt',
        data_prefix=dict(img_path='GOT10k'),
        pipeline=test_pipeline,
        test_mode=True))
test_dataloader = val_dataloader

# evaluator
val_evaluator = dict(
    type='SOTMetric',
    format_only=True,
    metric_options=dict(dataset_type='got10k'))
test_evaluator = val_evaluator

# runner loop
val_cfg = dict(type='ValLoop')
test_cfg = dict(type='TestLoop')



# Copyright (c) OpenMMLab. All rights reserved.
import os.path as osp
import random
from abc import ABCMeta
from io import StringIO
from typing import Any, Optional, Sequence, Union

import numpy as np
from addict import Dict
from mmengine.dataset import BaseDataset, force_full_init
from mmengine.fileio.file_client import FileClient

from mmtrack.registry import DATASETS


@DATASETS.register_module()
class BaseSOTDataset(BaseDataset, metaclass=ABCMeta):
    """Base dataset for SOT task. The dataset can both support training and
    testing mode.

    Args:
        bbox_min_size (int, optional): Only bounding boxes whose sizes are
            larger than ``bbox_min_size`` can be regarded as valid.
            Default to 0.
        only_eval_visible (bool, optional): Whether to only evaluate frames
            where object are visible. Default to False.
    """

    META = dict(CLASSES=None)

    def __init__(self,
                 bbox_min_size: int = 0,
                 only_eval_visible: bool = False,
                 *args,
                 **kwargs):
        self.bbox_min_size = bbox_min_size
        self.only_eval_visible = only_eval_visible
        # ``self.load_as_video`` must be set to True in order to using
        # distributed video sampler to load dataset when testing.
        self.load_as_video = True
        super().__init__(*args, **kwargs)

        # used to record the video information at the beginning of the video
        # test. Thus, we can avoid reloading the files of video information
        # repeatedly in all frames of one video.
        self.test_memo = Dict()

    def _loadtxt(self,
                 filepath: str,
                 dtype=np.float32,
                 delimiter: Optional[str] = None,
                 skiprows: int = 0,
                 return_ndarray: bool = True) -> Union[np.ndarray, str]:
        """Load TEXT file.

        Args:
            filepath (str): The path of file.
            dtype (data-type, optional): Data-type of the resulting array.
                Defaults to np.float32.
            delimiter (str, optional): The string used to separate values.
                Defaults to None.
            skiprows (int, optional): Skip the first ``skiprows`` lines,
                including comments. Defaults to 0.
            return_ndarray (bool, optional): Whether to return the ``ndarray``
                type. Defaults to True.

        Returns:
            Union[np.ndarray, str]: Contents of the file.
        """
        file_client = FileClient.infer_client(uri=filepath)
        file_string = file_client.get_text(filepath)
        if return_ndarray:
            return np.loadtxt(
                StringIO(file_string),
                dtype=dtype,
                delimiter=delimiter,
                skiprows=skiprows)
        else:
            return file_string.strip()

    def get_bboxes_from_video(self, video_idx: int) -> np.ndarray:
        """Get bboxes annotation about the instance in a video.

        Args:
            video_idx (int): video index

        Returns:
            np.ndarray: In [N, 4] shape. The N is the number of bbox and
                the bbox is in (x, y, w, h) format.
        """
        meta_video_info = self.get_data_info(video_idx)
        bbox_path = osp.join(self.data_prefix['img_path'],
                             meta_video_info['ann_path'])
        bboxes = self._loadtxt(bbox_path, dtype=float, delimiter=',')
        if len(bboxes.shape) == 1:
            bboxes = np.expand_dims(bboxes, axis=0)

        end_frame_id = meta_video_info['end_frame_id']
        start_frame_id = meta_video_info['start_frame_id']

        if not self.test_mode:
            assert len(bboxes) == (
                end_frame_id - start_frame_id + 1
            ), f'{len(bboxes)} is not equal to {end_frame_id}-{start_frame_id}+1'  # noqa
        return bboxes

    def get_len_per_video(self, video_idx: int) -> int:
        """Get the number of frames in a video.

        Args:
            video_idx (int): The index of video.

        Returns:
            int: The length of the video.
        """
        return self.get_data_info(
            video_idx)['end_frame_id'] - self.get_data_info(
                video_idx)['start_frame_id'] + 1

    def get_visibility_from_video(self, video_idx: int) -> dict:
        """Get the visible information of instance in a video.

        Args:
            video_idx (int): The index of video.

        Returns:
            dict: The visibilities of each object in the video.
        """
        visible = np.array([True] * self.get_len_per_video(video_idx))
        return dict(visible=visible)

    def get_masks_from_video(self, video_idx: int) -> Any:
        """Get the mask information of instance in a video.

        Args:
            video_idx (int): The index of video.

        Returns:
            Any: Not implemented yet.
        """
        pass

    def get_img_infos_from_video(self, video_idx: int) -> dict:
        """Get the information of images in a video.

        Args:
            video_idx (int): The index of video.

        Returns:
            dict: {
                    'video_id': int,
                    'frame_ids': np.ndarray,
                    'img_paths': list[str],
                    'video_length': int
                  }
        """
        img_paths = []
        meta_video_info = self.get_data_info(video_idx)
        start_frame_id = meta_video_info['start_frame_id']
        end_frame_id = meta_video_info['end_frame_id']
        framename_template = meta_video_info['framename_template']
        for frame_id in range(start_frame_id, end_frame_id + 1):
            img_paths.append(
                osp.join(self.data_prefix['img_path'],
                         meta_video_info['video_path'],
                         framename_template % frame_id))
        video_len = self.get_len_per_video(video_idx)
        frame_ids = np.arange(video_len)

        img_infos = dict(
            video_id=video_idx,
            frame_ids=frame_ids,
            img_paths=img_paths,
            video_length=video_len)
        return img_infos

    def get_ann_infos_from_video(self, video_idx: int) -> dict:
        """Get the information of annotations in a video.

        Args:
            video_idx (int): The index of video.

        Returns:
            dict: {
                    'bboxes': np.ndarray in (N, 4) shape,
                    'bboxes_isvalid': np.ndarray,
                    'visible': np.ndarray
                  }.
                  The annotation information in some datasets may contain
                    'visible_ratio'. The bbox is in (x1, y1, x2, y2) format.
        """
        bboxes = self.get_bboxes_from_video(video_idx)
        # The visible information in some datasets may contain
        # 'visible_ratio'.
        visible_info = self.get_visibility_from_video(video_idx)
        bboxes_isvalid = (bboxes[:, 2] > self.bbox_min_size) & (
            bboxes[:, 3] > self.bbox_min_size)
        visible_info['visible'] = visible_info['visible'] & bboxes_isvalid
        bboxes[:, 2:] += bboxes[:, :2]

        ann_infos = dict(
            bboxes=bboxes, bboxes_isvalid=bboxes_isvalid, **visible_info)
        return ann_infos

    def prepare_test_data(self, video_idx: int, frame_idx: int) -> dict:
        """Get testing data of one frame. We parse one video, get one frame
        from it and pass the frame information to the pipeline.

        Args:
            video_idx (int): The index of video.
            frame_idx (int): The index of frame.

        Returns:
            dict: Testing data of one frame.
        """
        # Avoid reloading the files of video information
        # repeatedly in all frames of one video.
        if self.test_memo.get('video_idx', None) != video_idx:
            self.test_memo.video_idx = video_idx
            ann_infos = self.get_ann_infos_from_video(video_idx)
            img_infos = self.get_img_infos_from_video(video_idx)
            self.test_memo.video_infos = dict(**img_infos, **ann_infos)
        assert 'video_idx' in self.test_memo and 'video_infos'\
            in self.test_memo

        results = {}
        results['img_path'] = self.test_memo.video_infos['img_paths'][
            frame_idx]
        results['frame_id'] = frame_idx
        results['video_id'] = video_idx
        results['video_length'] = self.test_memo.video_infos['video_length']

        results['instances'] = []
        instance = {}
        instance['bbox'] = self.test_memo.video_infos['bboxes'][frame_idx]
        instance['visible'] = self.test_memo.video_infos['visible'][frame_idx]
        instance['bbox_label'] = np.array([0], dtype=np.int32)
        results['instances'].append(instance)

        results = self.pipeline(results)
        return results

    def prepare_train_data(self, video_idx: int) -> dict:
        """Get training data sampled from some videos. We firstly sample two
        videos from the dataset and then parse the data information in the
        subsequent pipeline. The first operation in the training pipeline must
        be frames sampling.

        Args:
            video_idx (int): The index of video.

        Returns:
            dict: Training data pairs, triplets or groups.
        """
        video_idxes = random.choices(list(range(len(self))), k=2)
        pair_video_infos = []
        for video_idx in video_idxes:
            ann_infos = self.get_ann_infos_from_video(video_idx)
            img_infos = self.get_img_infos_from_video(video_idx)
            video_infos = dict(**img_infos, **ann_infos)
            pair_video_infos.append(video_infos)

        results = self.pipeline(pair_video_infos)
        return results

    def prepare_data(self, idx: Union[Sequence[int], int]) -> Any:
        """Get data processed by ``self.pipeline``.

        Args:
            idx (int): The index of ``data_info``.

        Returns:
            Any: Depends on ``self.pipeline``.
        """
        if self.test_mode:
            assert isinstance(idx, Sequence) and len(idx) == 2
            # the first element in the ``Sequence`` is the video index and the
            # second element in the ``Sequence`` is the frame index
            return self.prepare_test_data(idx[0], idx[1])
        else:
            assert isinstance(idx, int)
            return self.prepare_train_data(idx)

    @property
    def num_videos(self) -> int:
        """Get the number of videos in the dataset.

        Returns:
            int: The number of videos.
        """
        num_videos = len(self.data_address) if self.serialize_data else len(
            self.data_list)
        return num_videos

    @force_full_init
    def __len__(self) -> int:
        """Get the length of filtered dataset and automatically call
        ``full_init`` if the  dataset has not been fully init.

        Returns:
            int: The length of filtered dataset.
        """
        if self.test_mode:
            return sum(
                self.get_len_per_video(idx) for idx in range(self.num_videos))
        else:
            return self.num_videos



# Copyright (c) OpenMMLab. All rights reserved.
import math
from typing import Optional, Tuple

import torch
import torch.nn as nn
from mmengine.model import BaseModule
from torch import Tensor

from mmtrack.registry import MODELS
from ..task_modules.filter import filter as filter_layer


@MODELS.register_module()
class PrDiMPFilterOptimizer(BaseModule):
    """Optimizer module of filter in PrDiMP.

    It unrolls the steepest descent with Newton iterations to optimize the
        target filter.

    Args:
        num_iters (int, optional):  Number of default optimization iterations.
            Defaults to 1.
        feat_stride (int, optional):  The stride of the input feature.
            Defaults to 16.
        init_step_length (float, optional):  Initial scaling of the step length
            (which is then learned). Defaults to 1.0.
        init_filter_regular (float, optional):  Initial filter regularization
            weight (which is then learned). Defaults to 1e-2.
        gauss_sigma (float, optional):  The standard deviation to use for the
            label density function. Defaults to 1.0.
        min_filter_regular (float, optional):  Enforce a minimum value on the
            regularization (helps stability sometimes). Defaults to 1e-3.
        alpha_eps (float, optional):  Term in the denominator of the steepest
            descent that stabalizes learning. Defaults to 0.
        label_thres (float, optional):  Threshold probabilities. Defaults to 0.
    """

    def __init__(self,
                 num_iters: int = 1,
                 feat_stride: int = 16,
                 init_step_length: float = 1.0,
                 init_filter_regular: float = 1e-2,
                 gauss_sigma: float = 1.0,
                 min_filter_regular: float = 1e-3,
                 alpha_eps: float = 0.0,
                 label_thres: float = 0.0):
        super().__init__()

        self.num_iters = num_iters
        self.feat_stride = feat_stride
        self.log_step_length = nn.Parameter(
            math.log(init_step_length) * torch.ones(1))
        self.filter_regular = nn.Parameter(init_filter_regular * torch.ones(1))
        self.gauss_sigma = gauss_sigma
        self.min_filter_regular = min_filter_regular
        self.alpha_eps = alpha_eps
        self.label_thres = label_thres

    def gen_label_density(self, center_yx: Tensor,
                          output_size_hw: Tensor) -> Tensor:
        """Generate label density.

        Args:
            center_yx (Tensor): The center of score map.
            output_size_hw (Tensor): The size of score map in [h, w] format.

        Returns:
            Tensor: Label density with two possible shape:
                - train mode: (num_img_per_seq, bs, h, w).
                - test mode: (num_img_per_seq, 1, h, w).
        """
        # convert to (num_img_per_seq, bs, 4) shape
        center_yx = center_yx.reshape(center_yx.shape[0], -1,
                                      center_yx.shape[-1])
        k0 = torch.arange(
            output_size_hw[0],
            dtype=torch.float32).reshape(1, 1, -1, 1).to(center_yx.device)
        k1 = torch.arange(
            output_size_hw[1],
            dtype=torch.float32).reshape(1, 1, 1, -1).to(center_yx.device)
        dist0 = (k0 -
                 center_yx[:, :, 0].reshape(*center_yx.shape[:2], 1, 1))**2
        dist1 = (k1 -
                 center_yx[:, :, 1].reshape(*center_yx.shape[:2], 1, 1))**2
        if self.gauss_sigma == 0:
            dist0_view = dist0.reshape(-1, dist0.shape[-2])
            dist1_view = dist1.reshape(-1, dist1.shape[-1])
            one_hot0 = torch.zeros_like(dist0_view)
            one_hot1 = torch.zeros_like(dist1_view)
            one_hot0[torch.arange(one_hot0.shape[0]),
                     dist0_view.argmin(dim=-1)] = 1.0
            one_hot1[torch.arange(one_hot1.shape[0]),
                     dist1_view.argmin(dim=-1)] = 1.0
            gauss = one_hot0.reshape(dist0.shape) * one_hot1.reshape(
                dist1.shape)
        else:
            g0 = torch.exp(-1.0 / (2 * self.gauss_sigma**2) * dist0)
            g1 = torch.exp(-1.0 / (2 * self.gauss_sigma**2) * dist1)
            gauss = (g0 / (2 * math.pi * self.gauss_sigma**2)) * g1
        gauss = gauss * (gauss > self.label_thres).float()
        gauss_density = gauss / (gauss.sum(dim=(-2, -1), keepdim=True) + 1e-8)
        return gauss_density

    def forward(self,
                filter_weights: Tensor,
                feat: Tensor,
                bboxes: Tensor,
                num_iters: Optional[int] = None,
                sample_weights: Optional[Tensor] = None) -> Tuple[Tensor, ...]:
        """Runs the optimizer module.

        Note that [] denotes an optional dimension. Generally speaking, inputs
        in test mode don't have the dim of [].

        Args:
            filter_weights (Tensor):  Initial filter with shape
                training mode: (bs, c, fitler_h, filter_w)
                test mode: (1, c, fitler_h, filter_w)
            feat (Tensor):  Input feature maps with shape
                (num_img_per_seq, [bs], c, H, W).
            bboxes (Tensor):  Target bounding boxes with shape
                (num_img_per_seq, [bs], 4). in (cx, cy, x, y) format.
            num_iters (int, optional):  Number of iterations to run.
                Defaults to None.
            sample_weights (Tensor, optional):  Optional weight for each
                sample with shape (num_img_per_seq, [bs]). Defaults to None.

        Returns:
            filter_weights (Tensor):  The final oprimized filter.
            filter_iters (Tensor, optional):  The filter computed in each
                iteration (including initial input and final output), returned
                only in training
            losses (Tensor, optional): losses in all optimizer iterations,
                returned only in training
        """

        # Sizes
        num_iters = self.num_iters if num_iters is None else num_iters
        num_img_per_seq = feat.shape[0]
        batch_size = feat.shape[1] if feat.dim() == 5 else 1
        filter_size_hw = (filter_weights.shape[-2], filter_weights.shape[-1])
        output_size_hw = (feat.shape[-2] + (filter_weights.shape[-2] + 1) % 2,
                          feat.shape[-1] + (filter_weights.shape[-1] + 1) % 2)

        # Get learnable scalars
        step_length_factor = torch.exp(self.log_step_length)
        filter_regular = (self.filter_regular**2).clamp(
            min=self.min_filter_regular**2)

        # Compute label density
        if self.training:
            assert bboxes.dim() == 3
        else:
            assert bboxes.dim() == 2
            bboxes = bboxes.reshape([bboxes.shape[0], -1, bboxes.shape[-1]])
        offset = (torch.Tensor(filter_size_hw).to(bboxes.device) % 2) / 2.0
        center = bboxes[..., :2] / self.feat_stride
        center_yx = center.flip((-1, )) - offset
        label_density = self.gen_label_density(center_yx, output_size_hw)

        # Get total sample weights
        if sample_weights is None:
            sample_weights = torch.Tensor([1.0 / num_img_per_seq
                                           ]).to(feat.device)
        elif isinstance(sample_weights, torch.Tensor):
            sample_weights = sample_weights.reshape(num_img_per_seq,
                                                    batch_size, 1, 1)
        else:
            raise NotImplementedError(
                "Only support two types of 'sample_weights': "
                'torch.Tensor or None')

        filter_iters = []
        losses = []

        for _ in range(num_iters):
            # Get scores by applying the filter on the features
            scores = filter_layer.apply_filter(feat, filter_weights)
            scores = torch.softmax(
                scores.reshape(num_img_per_seq, batch_size, -1),
                dim=2).reshape(scores.shape)

            # Compute loss and record the filter of each iteration in training
            # mode.
            if self.training:
                filter_iters.append(filter_weights)
                losses.append(
                    self._compute_loss(scores, sample_weights, label_density,
                                       filter_weights, filter_regular))

            # Compute gradient and step_length
            res = sample_weights * (scores - label_density)
            filter_grad = filter_layer.apply_feat_transpose(
                feat, res, filter_size_hw,
                training=self.training) + filter_regular * filter_weights

            step_length = self.get_step_length(feat, sample_weights, scores,
                                               filter_grad, filter_regular)

            # Update filter
            filter_weights = filter_weights - (
                step_length_factor *
                step_length.reshape(-1, 1, 1, 1)) * filter_grad

        if self.training:
            filter_iters.append(filter_weights)
            # Get scores by applying the final filter on the feature map
            scores = filter_layer.apply_filter(feat, filter_weights)
            losses.append(
                self._compute_loss(scores, sample_weights, label_density,
                                   filter_weights, filter_regular))
            return filter_weights, filter_iters, losses
        else:
            return filter_weights

    def get_step_length(self, feat: Tensor, sample_weights: Tensor,
                        scores: Tensor, filter_grad: Tensor,
                        filter_regular: Tensor) -> Tensor:
        """Compute the step length of updating the filter.

        Args:
            feat (Tensor): Input feature map with shape
                (num_img_per_seq, [bs], feat_dim, H, W).
            sample_weights (Tensor): The weights of all the samples.
            scores (Tensor): The score map with two possible shape:
                - train mode: (num_img_per_seq, bs, h, w).
                - test mode: (num_img_per_seq, 1, h, w).
            filter_grad (Tensor): The gradient of the filter with shape
                (num_img_per_seq, c, fitler_h, filter_w).
            filter_regular (Tensor): The regulazation item of the filter, with
                shape (1,).

        Returns:
            alpha (Tensor): The updating factor with shape (1, ).
        """
        num_img_per_seq = feat.shape[0]
        batch_size = feat.shape[1] if feat.dim() == 5 else 1
        # Map the gradient with the Hessian
        scores_grad = filter_layer.apply_filter(feat, filter_grad)
        sm_scores_grad = scores * scores_grad
        hes_scores_grad = sm_scores_grad - scores * torch.sum(
            sm_scores_grad, dim=(-2, -1), keepdim=True)
        grad_hes_grad = (scores_grad * hes_scores_grad).reshape(
            num_img_per_seq, batch_size, -1).sum(dim=2).clamp(min=0)
        grad_hes_grad = (sample_weights.reshape(sample_weights.shape[0], -1) *
                         grad_hes_grad).sum(dim=0)

        # Compute optimal step length
        alpha_num = (filter_grad * filter_grad).sum(dim=(1, 2, 3))
        alpha_den = (grad_hes_grad +
                     (filter_regular + self.alpha_eps) * alpha_num).clamp(1e-8)
        alpha = alpha_num / alpha_den

        return alpha

    def _compute_loss(self, scores: Tensor, sample_weights: Tensor,
                      label_density: Tensor, filter: Tensor,
                      filter_regular: Tensor) -> Tensor:
        """Compute loss in the box optimization.

        Args:
            scores (Tensor): The score map with shape
                (num_img_per_seq, bs, h, w).
            sample_weights (Tensor): The weights of all the samples with shape
                (num_img_per_seq, bs, 1, 1)
            label_density (Tensor):The label density with shape
                (num_img_per_seq, bs, h, w).
            filter (Tensor): The filter with shape
                (num_img_per_seq, c, fitler_h, filter_w).
            filter_regular (Tensor):The regulazation item of the filter, with
                shape (1,).

        Returns:
            Tensor: with shape (1,)
        """

        num_samples = sample_weights.shape[0]
        sample_weights = sample_weights.reshape(sample_weights.shape[0], -1)
        score_log_sum_exp = torch.log(scores.exp().sum(dim=(-2, -1)))
        sum_scores = (label_density * scores).sum(dim=(-2, -1))

        return torch.sum(
            sample_weights * (score_log_sum_exp - sum_scores)
        ) / num_samples + filter_regular * (filter**2).sum() / num_samples
